<?php

/**
 * @file
 * Theme preprocessor functions.
 */
use Drupal\mcapi\Functions;
use Drupal\user\Entity\User;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Element;

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mc_wallet(&$vars) {
  $vars['canonical'] = $vars['element']['#mc_wallet']->toUrl('canonical')->toString();
  $vars['attributes']['class'] = ['wallet', $vars['element']['#view_mode']];
}

/**
 * Theme preprocessor for 'wallet_component'.
 */
function template_preprocess_mc_wallet_component(&$vars) {
  $vars['title'] = @$vars['element']['#title'];
  $vars['attributes']['class'][] = 'component';
  $vars['attributes']['class'][] = str_replace('_', '-', substr($vars['element']['#theme'], 7));
}

/**
 * Theme preprocessor for all of an entity's wallets.
 */
function template_preprocess_mc_wallets(&$vars) {
  $viewbuilder = \Drupal::entityTypeManager()->getViewBuilder('mc_wallet');
  // Convert each wallet into a render array.
  foreach ($vars['wallets'] as $key => $wallet) {
    $vars['wallets'][$key] = $viewbuilder->view($wallet, $vars['view_mode']);
  }
}

/**
 * Theme preprocessor for 'balance_bars'.
 *
 * All values sent to the theme system as numbers should be formatted plainly,
 * all labels formatted normally.
 */
function template_preprocess_mc_wallet_balance_bars(&$vars) {
  $wallet = $vars['wallet'];
  $stats = $wallet->getAllStats();
  $id = "given-received" . $wallet->id();
  $max_raw = max($stats->gross_in, $stats->gross_out);
  $max_formatted_plain = Functions::formatCurrencyAsFloat(mcapi_max_axis($max_raw));
  $vars += [
    //'label' => t('Reciprocation'),
    'id' => $id,
    'functionname' => str_replace('-', '_', $id),
    'incoming' => Functions::formatCurrencyAsFloat($stats->gross_in),
    'outgoing' => Functions::formatCurrencyAsFloat($stats->gross_out),
    'show_in' => Functions::formatCurrency($stats->gross_in),
    'show_out' => Functions::formatCurrency($stats->gross_out),
    'max' => $max_formatted_plain,
    'vaxislabels' => [
      [
        'value' => 0,
        'label' => strip_tags(Functions::formatCurrency(0)),
      ],
      [
        'value' => floor($max_formatted_plain / 2),
        'label' => strip_tags(Functions::formatCurrency(intval($max_raw / 2))),
      ],
      [
        'value' => $max_formatted_plain,
        'label' => strip_tags(Functions::formatCurrency($max_raw)),
      ],
    ],
  ];
}

/**
 * Theme preprocessor for 'balance_viz'.
 */
function template_preprocess_mc_wallet_balance_viz(&$vars) {
  $stats = $vars['wallet']->getAllStats();
  if ($stats->gross_in > $stats->gross_out) {
    $vars['side'] = 'positive';
  }
  elseif ($stats->gross_in < $stats->gross_out) {
    $vars['side'] = 'negative';
  }
  else {
    $vars['side'] = 'balanced';
  }
  $vars['income_raw'] = $stats->gross_in;
  $vars['expenditure_raw'] = $stats->gross_out;
  $vars['income'] = Markup::create($vars['wallet']->gross_in);
  $vars['expenditure'] = Markup::create($vars['wallet']->gross_out);
  $vars['balance'] = Markup::create($vars['wallet']->balance);
}

/**
 * Theme preprocessor for 'mc_wallet_stats'.
 */
function template_preprocess_mc_wallet_stats(&$vars) {
  $data = $vars['wallet']->getAllStats();
  $vars['gross_in'] = Functions::formatCurrency($data->gross_in);
  $vars['gross_out'] = Functions::formatCurrency($data->gross_out);
  $vars['volume'] = Functions::formatCurrency($data->volume);
  $vars['balance'] = Functions::formatCurrency($data->balance);
  $vars['partners'] = $data->partners;
  $vars['trades'] = $data->trades;
}


/**
 * Theme preprocessor for 'mc_main_entry'.
 */
function template_preprocess_mc_main_entry(&$vars) {
  $elements = $vars['elements'];
  $vars['payee'] = $elements['#mc_entry']->payee->label();
  $vars['payer'] = $elements['#mc_entry']->payer->label();
  $vars['description'] = $elements['#mc_entry']->description;
  $quant = Functions::formatCurrency($elements['#mc_entry']->quant);
  $vars['quant'] = Markup::create('<div class="mc-quantity">'.$quant.'</div>');
}

/**
 * Theme preprocessor for 'mc_transaction'.
 */
function template_preprocess_mc_transaction(&$vars) {
  $elements = $vars['elements'];
  $renderer = \Drupal::service('renderer');
  // Render the first and additional entries into separate variables.
  $vars['entry_main'] = $renderer->render($elements['entries'][0]);
  $vars['entries_other'] = $renderer->render($elements['entries'][1]);
  unset($elements['entries']);// crude

  $vars['signatures'] = $renderer->render($elements['signatures']);
  $vars['created'] = $renderer->render($elements['created'][0]);
  $other_fields = [];
  foreach (Element::children($elements) as $key) { // unrendered fields
    if (!isset($vars[$key])) {
      $other_fields[$key] = $renderer->render($vars['elements'][$key]);
    }
  }
  $vars['other_fields'] = Markup::create(implode($other_fields));
  if (!empty($elements['#watermark'])) {
    // To add a watermark to the render array, add #watermark => blah
    $vars['watermark'] = Markup::create('<div class = "watermark">'.$elements['#watermark'].'</div>');
  }
  if (isset($elements['#links'])) {
    $vars['links'] = $renderer->render($elements['#links']);
  }
}

/**
 * Theme preprocessor for 'mc_timeline'.
 *
 * Adds a column to the data with the formatted currency value.
 */
function template_preprocess_mc_timeline(&$vars) {
  $vars['id'] = preg_replace('/[^a-z0-9_]+/', '_', strtolower($vars['title']));
  $points = &$vars['points'];
  $points[\Drupal::time()->getRequestTime()] = end($points);

  // Apply smoothing, or even roughing.
  if (count($points) < $vars['width'] / 3) {
    // Step method, for a small number of points.
    $times = $values = [];
    // Make two values for each one in the keys and values.
    foreach ($points as $time => $bal) {
      $times[] = $time;
      $times[] = $time + 1;
      $values[] = $bal;
      $values[] = $bal;
    }
    // Now slide the arrays against each other to create steps.
    array_pop($values);
    array_shift($times);
    $points = array_combine($times, $values);
  }
  elseif (count($points) > $vars['width']) {
    // Decimate the array, for a large number of points.
    $ratio = count($points) / $vars['width'];
    $factor = intval($ratio + 1);
    // Now iterate through the array taking 1 out of every $factor values.
    $i = 0;
    foreach ($points as $key => $value) {
      if ($i % $factor != 0) {
        unset($points[$key]);
      }
      $i++;
    }
  }
  // Prepare for javascript
  foreach ($points as $timestamp => $val) {
    $vars['daterows'][date('Y,m,d', $timestamp)] = $val;
  }
  // TEMP until gcharts works better. Force some formatted values on the axis.
  list($min, $mid, $max) = mcapi_gchart_axes($vars['points']);
  $vars['vaxislabels'] = [
    ['value' => $min, 'label' => $min],
    ['value' => $mid, 'label' => $mid],
    ['value' => $max, 'label' => $max],
  ];
  // @todo is there a way to have html labels in google charts?
  if ($vars['vaxislabels']) {
    foreach ($vars['vaxislabels'] as &$tick) {
      $tick['label'] = strip_tags($tick['label']);
    }
  }
  if ($vars['format_curr']) {
    $vars['daterows'] = array_map(function($v){return Functions::formatCurrencyAsFloat($v);}, $vars['daterows']);
    foreach ($vars['vaxislabels'] as &$lab) {
      $lab['value'] = Functions::formatCurrencyAsFloat($lab['value']);
      $lab['label'] = strip_tags(Functions::formatCurrencyAsFloat($lab['value'], FALSE));
    }
  }
}

/**
 * Implements template_preprocess_THEMEHOOK().
 * @note this is probably renamed.
 */
function template_preprocess_mc_signatures(&$vars) {
  foreach ($vars['signatures'] as $delta => $sig) {
    $user_link = User::load($sig['uid'])->toLink();
    $vars['items'][$delta]['account'] = Markup::create($user_link->toString());
    if (!empty($sig['signed'])) {
      $vars['items'][$delta]['date'] = \Drupal::service('date.formatter')->format($sig['signed']);
    }
  }
}


/**
 * Calculate some good axis labels, using the min & max balance extents.
 *
 * @param array $vals
 *   The $points history of raw balances keyed by unixtimes.
 *
 * @return int[]
 *   Integers keyed by min, 0 & max.
 */
function mcapi_gchart_axes(array $vals) : array {
  $max = $vals ? max($vals) : 0;
  $min = $vals ? min($vals) : 0;
  if ($min >= 0) {
    $max = mcapi_max_axis($max);
    return [0, intval($max / 2), $max];
  }
  elseif ($max <= 0) {
    $min = -mcapi_max_axis(abs($min));
    return [$min, intval($min / 2), 0];
  }
  else {
    return [-mcapi_max_axis(abs($min)), 0, mcapi_max_axis($max)];
  }
}

/**
 * Helper.
 *
 * @param int $val
 *   The max value to represent in the chart.
 */
function mcapi_max_axis($val) : int {
  $max = 10 ** ceil(log($val, 10));
  while ($max/2 > $val) {
    $max = $max / 2;
  }
  return intval($max);
}
