<?php

namespace Drupal\mcapi_credcom\Form;

use CreditCommons\Exceptions\CCFailure;
use Drupal\mcapi_credcom\Functions;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Workflow;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Extension\ModuleHandler;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;


class SettingsForm extends ConfigFormBase {

  protected $config;
  private bool $traded = FALSE;
  protected bool $connected = FALSE;
  protected ModuleHandler $moduleHandler;

  /**
   * {@inheritDoc}
   */
  public function getFormID() {
    return 'mcapi_credcom_settings_form';
  }

  /**
   *
   * @param ConfigFactory $config_factory
   * @param EntityTypeManager $entity_type_manager
   * @param ModuleHandler $module_handler
   */
  public function __construct(ConfigFactory $config_factory, EntityTypeManager $entity_type_manager, ModuleHandler $module_handler) {
    parent::__construct($config_factory);
    $this->config = $config_factory->get('mcapi_credcom.settings');
    $this->traded = (bool)$entity_type_manager->getStorage('mc_transaction')->getQuery()
      ->accessCheck(FALSE)
      ->involving($this->config->get('wallet_id'))
      ->execute();
    $this->moduleHandler = $module_handler;
    Functions::setErrorContext($this->config);

  }

  static public function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $w = 0;

    $url = $this->config->get('trunkward.url');
    if ($ccuser = $this->config->get('node_name') and $url) {
      try {
        $this->connected = $this->handshake($url, $ccuser);
      }
      catch(\CreditCommons\Exceptions\CCError $e) {
        \Drupal::messenger()->addStatus($e->getMessage());
      }
      catch(\Exception $e) {}
    }

    $form['node_name'] = [
      '#title' => $this->t('The name of the current site on the network'),
      '#description' => $this->config->get('node_name') ? t('This cannot be changed.') : t('The account must exist already on the trunkwards node.'),
      '#type' => 'textfield',
      '#default_value' => $ccuser,
      '#placeholder' => $this->t('lowercase with _ and no spaces.'),
      '#pattern' => '[a-z1-9_]+',
      '#weight' => $w++,
      '#disabled'=> isset($connected),
      '#required' => TRUE
    ];
    $form['wallet_name'] = [
      '#title' => $this->t('Friendly name of wallet @num (the intertrading wallet).', ['@num' => $this->config->get('wallet_id')]),
      '#type' => 'textfield',
      '#default_value' => $this->config->get('wallet_name'),
      '#weight' => $w++,
      '#required' => TRUE
    ];
    $form['trunkward'] = [
      '#title' => $this->t('Connection to trunkwards node'),
      '#type' => 'details',
      '#open' => !$this->connected,
      '#tree' => TRUE,
      '#weight' => $w++,
      'id' => [
        '#title' => $this->t('Trunkwards node ID'),
        '#description' => $this->t('The cc_user name the trunkward node uses to connect'),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('trunkward.id'),
        '#pattern' => '[a-z1-9]+',
        '#required' => TRUE,
        '#disabled'=> $this->connected,
        '#weight' => 1
      ],
      'url' => [
        '#title' => $this->t('Trunkward node URL'),
        '#description' => $this->t('Must be https to protect the login credentials'),
        '#type' => 'textfield',
        '#default_value' => $url,
        '#placeholder' => 'https://',
        '#pattern' => 'https?:\/\/[a-z/.\-0-9]+', // https enforced except during development
        '#disabled'=> $this->connected,
        '#required' => TRUE,
        '#weight' => 2
      ],
      'rate' => [
        '#title' => $this->t('Conversion rate'),
        '#description' => $this->t('The number of (raw) local units to one unit on the parent ledger. Note that local units are stored as integers of the smallest subdivision, e.g. cents not dollars, minutes, not hours.'),
        '#type' => 'number',
        '#min' => 0,
        '#step' => 0.01,
        '#default_value' => $this->config->get('trunkward.rate'),
        '#disabled'=> $this->traded,
        '#weight' => 3,
      ],
      'symbol'=> [
        '#title' => $this->t('Symbol for the local currency'),
        '#description' => $this->t('An html string which is placed before the numeric value.'),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('trunkward.symbol'),
        '#weight' => 4,
        '#required' => TRUE
      ]
    ];
    $form['remote_forms'] = [
      '#title' => $this->t('Scope of forms using inherited workflows'),
      '#description' => $this->t('Forms for transactions using workflows that are shared with the trunkward node can be set to support local wallets, remote wallets, or either.'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
      '#weight' => $w++,
    ];
    $remote_forms = array_filter(
      $this->moduleHandler->invokeAll('mcapi_form_list'),
      function ($w, $k) {
        $wf = Workflow::load($w['workflow']);
        return $wf->disabled != TRUE and $wf->local == FALSE and strpos($k, 'mass') !== 0;
      },
      ARRAY_FILTER_USE_BOTH
    );
    if ($remote_forms) {
      foreach ($remote_forms as $form_id => $info) {
        $url = Url::fromRoute($info['route_name'], @$info['route_parameters'], ['attributes' => ['target' => '_blank']]);
        $form['remote_forms'][$form_id] = [
          '#title' => $info['label'],
          '#description' => Link::fromTextAndUrl(t('View form'), $url)->toString(),
          '#type' => 'select',
          '#options' => [
            CREDCOM_FORM_LOCAL => $this->t('Local only'),
            CREDCOM_FORM_BOTH => $this->t('Local OR remote'),
            CREDCOM_FORM_REMOTE => $this->t('Remote only')
          ],
          '#default_value' => $this->config->get('remote_forms.'.$form_id) ?? CREDCOM_FORM_BOTH
        ];
      }
    }
    else {
      $form['remote_forms']['empty']['#markup'] = t('No remote workflows until the trunkward node is connected.');
    }
    $form['privacy'] = [
      '#title' => $this->t('Privacy'),
      '#description' => $this->t('This site is a leaf on a Credit Commons tree. What data can be shared with the rest of the tree?'),
      '#type' => 'details',
      '#open' => TRUE,
      '#tree' => TRUE,
      '#weight' => $w++,
      'acc_path_autocomplete' => [
        '#title' => $this->t('Wallet autocomplete'),
        '#description' => $this->t('Incoming transactions can address wallets using the name, with dashes instead of spaces.').' '.$this->t('How much personal information can be seen by traders in other nodes trying to autocomplete wallet addresses?'),
        '#type' => 'select',
        '#options' => [
          CREDCOM_AUTOCOMPLETE_NONE => $this->t('Disable autocomplete'),
          CREDCOM_AUTOCOMPLETE_NUM => $this->t('Numeric wallet ids'),
          CREDCOM_AUTOCOMPLETE_NAME => $this->t('Expose wallet names'),
        ],
        '#default_value' => $this->config->get('privacy.acc_path_autocomplete'),
        '#required' => TRUE,
      ],
      'aggregate_stats' => [
        '#title' => $this->t('Expose aggregate trading statistics') . ' NOT IMPLEMENTED',
        '#description' => $this->t('Share trading statistics with the Credit commons network.'),
        '#type' => 'checkbox',
        '#default_value' => $this->config->get('privacy.aggregate_stats'),
        '#disabled' => TRUE
      ]
    ];

    return parent::buildForm($form, $form_state);
  }

  public function validateForm(array &$form, FormStateInterface $form_state) {
    try {
      $this->handshake(
        $form_state->getValue(['trunkward', 'url']),
        $form_state->getValue('node_name')
      );
    }
    catch (CCFailure $ex) {
      $form_state->setErrorByName('trunkward', $ex->getMessage());
    }
  }

  /**
   *
   * @param type $url
   * @param type $node_name
   * @throws CCFailure
   */
  protected function handshake(string $url, string $node_name) {
    $temp_requester = new \CreditCommons\NodeRequester($url, $node_name, Functions::getLastHash());
    $temp_requester->handshake();
    return TRUE;
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->configFactory->getEditable('mcapi_credcom.settings')
      ->set('node_name', $form_state->getValue('node_name'))
      ->set('trunkward', $form_state->getValue('trunkward'))
      ->set('wallet_name', $form_state->getValue('wallet_name'))
      ->set('remote_forms', $form_state->getValue('remote_forms'))
      ->set('privacy', $form_state->getValue('privacy'))
      ->save();

    Wallet::load($this->configFactory->get('mcapi_credcom.settings')
      ->get('wallet_id'))
      ->set('name', $form_state->getValue('wallet_name'))
      ->save();
    parent::submitForm($form, $form_state);
    \Drupal::service('mcapi_credcom.trunkward_node')->updateWorkflows();
  }

  /**
   * {@inheritDoc}
   */
  protected function getEditableConfigNames() {
    return ['mcapi_credcom.settings'];
  }

}
