<?php

namespace Drupal\mcapi_credcom;

use CreditCommons\BaseTransaction;
use CreditCommons\Exceptions\CCFailure;
use CreditCommons\Exceptions\CCError;
use CreditCommons\Exceptions\MaxLimitViolation;
use Drupal\mcapi\Entity\Workflow as McWorkflow;
use Drupal\mcapi_credcom\Functions;
use Drupal\mcapi_forms\Entity\McForm;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Render\Markup;
use GuzzleHttp\RequestOptions;


/**
 * Service handling outgoing requests to the trunkward node.
 */
class TrunkwardNodeRequester extends \CreditCommons\NodeRequester {

  private $logger;
  private $config;

  /**
   * Constructor
   *
   * @param ConfigFactory $config_factory
   * @param LoggerChannel $logger_channel
   */
  public function __construct(ConfigFactory $config_factory, LoggerChannel $logger_channel) {
    $this->config = $config_factory->get('mcapi_credcom.settings');
    Functions::setErrorContext($this->config);
    $this->logger = $logger_channel;
    if ($this->configured()) {
      // Initiate the service from the settings.
      parent::__construct(
        $this->config->get('trunkward.url'),
        $this->config->get('node_name'),
        Functions::getlastHash()
      );
    }
    $this->catch = TRUE;
  }

  public function configured() : bool {
    return $this->config->get('node_name') and $this->config->get('trunkward.url') and $this->config->get('trunkward.id');
  }

  /**
   * Ensure the local transaction->type reflect the remote workflows.
   *
   * @note workflows should change very rarely and even then should be
   * deprecated long before being removed to allow transactions to complete.
   * There is no procedure for converting transactions with old workflows to new.
   *
   * On update separate wfs into local only, remoteonly and both
   *   foreach localonly
   *     ensure the ->local is true.
   *   foreach remoteonly
   *     if a remote name is the same as a local name
   *       if the config->local is true
   *         rename the local config with local_ prefix && rewrite the ledger replacing the type name.
   *     copy the remote to config with ->local=false
   *   foreach both
   *     ensure the transactionType has local->false
   *     if the names are different
   *       rewrite the ledger to use the remote name.
   * @note this still allows that two workflows might have the same label
   */
  public function updateWorkflows() {
    $clear_cache = FALSE;
    [$remote_wfs, $local_wfs, $common_wfs] = Functions::compareWorkflows();
    foreach ($local_wfs as $local_wf) {
      /** @var \Drupal\mcapi\Entity\Type $local_wf */
      // All the items without trunkward equivalents should have the local flag set to TRUE
      $mcType = McWorkflow::load($local_wf->id);
      if (!$mcType->local) {
        $mcType->local = TRUE;
        $mcType->save();
        $clear_cache = TRUE;
        $this->logger->notice(
          'Workflow @name is no longer available remotely and is now considered a local workflow',
          ['@name' => $mcType->id]
        );
      }
      $labels[] = $local_wf->label;
    }
    foreach ($remote_wfs as $remote_wf) {
      /** @var CreditCommons\Workflow $remote_wf */
      // All trunkward workflows must be represented by local config.
      if ($mcType = McWorkflow::load($remote_wf->id)) {
        if (!$mcType->local) {
          $this->logger->error("Shouldn't be possible to have a remote workflow with and a different hash but the same name as a non-local transaction type.");
        }
        Functions::localiseTransactionTypeName($mcType); // Also updates existing transactions
        $this->logger->notice(
          'Workflow @name was the same as a remote workflow and has been renamed to @new_name',
          ['@name' => $remote_wf->id, '@new_name' => $mcType->id()]
        );
      }
      mcapi_credcom_convert_credcom_wf_to_mc_workflow($remote_wf)->save();
      $clear_cache = TRUE;
      $this->logger->notice(
        'The remote workflow @name has been saved locally',
        ['@name' => $remote_wf->id]
      );
        $clear_cache = TRUE;
      $labels[] = $remote_wf->label;
    }
    foreach ($common_wfs as $hash => $common_wf) {
      /** @var CreditCommons\Workflow $common_wf */
      // Common wfs must have the same name with the config->local set to false
      $mcType = McWorkflow::load($common_wf->localId);
      if ($common_wf->id <> $mcType->id) {
        Functions::updateTransactionTypes($mcType->id(), $common_wf->id);
        $map[$mcType->id()] = $common_wf->id;
        $mcType->delete();
        $clear_cache = TRUE;
        mcapi_credcom_convert_credcom_wf_to_mc_workflow($common_wf)->save();
        $this->logger->notice(
          'Replaced transactionType @type with remote workflow @id which had the same hash.',
          ['@type' => $mcType->id(), '@id' => $common_wf->id]
        );
      }
      elseif ($mcType->local) {
        $mcType->local = FALSE;
        $mcType->save();
        $clear_cache = TRUE;
        $this->logger->notice('Transaction type @type was converted to a local version of a remote workflow.');
      }
      $labels[] = $common_wf->label;
    }
    if (count($labels) > count(array_unique($labels))) {
      $this->logger->warning('Two workflows share the same label.');
    }
    if ($clear_cache) {
      \Drupal::service('cache_tags.invalidator')->invalidateTags(['entity_type:mc_workflow']);
    }
    // untested!
    // Change the workflow names in existing mc_forms
    if (class_exists('\Drupal\mcapi_forms\Entity\McForm')) {
      foreach (McForm::loadMultiple() as $form) {
        if (isset($map[$form->getWorkflowId()])) {
          $form->setWorkflowId($map[$form->getWorkflowId()])->save();
        }
      }
    }
  }

  /**
   * {@inheritDoc}
   *
   * Theoretically this could be disintermediated by javascript calling the
   * trunkward node directly, but then the creds would have to be sent to client.
   */
  public function accountNameFilter(string $acc_path = '', $limit = 10) : array {
    $result = [];
    try {
      $result = parent::accountNameFilter($acc_path, $limit);
    }
    catch (CCError $ex) {
      $this->logger->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->makeMessage(), '@fullerror'=> $ex]
      );
    }
    catch (\Throwable $ex) {
      $this->logger->error(
        "accountNameFilter returned @type @message<br />@fullerror",
        ['@type' => get_class($ex), '@message' => $ex->getMessage(), '@fullerror'=> $ex]
      );
    }
    return $result;
  }

  /**
   * {@inheritDoc}
   * @todo upgrade the API
   */
  public function convertPrice(string $node_path) : \stdClass {

  }

  /**
   * {@inheritDoc}
   */
  public function buildValidateRelayTransaction(BaseTransaction $transaction) : array {
    try {
      $additional = parent::buildValidateRelayTransaction($transaction);
    }
    catch (\CreditCommons\Exceptions\TransactionLimitViolation $e) {
      $rate = \Drupal::config('mcapi_credcom.settings')->get('trunkward.rate');
      // make a new exception that formats the value
      $args = [
        '@acc' => $e->acc_id,
        '@node' => $e->node,
        '@diff' => \Drupal\Mcapi\Functions::formatCurrency(round($e->diff / $rate), FALSE)
      ];
      // @todo how to render &#188; as 1/4 in the message box?
      if ($e instanceOf MaxLimitViolation) {
        $message = t("This transaction would put account '@acc' on node @node @diff above its maximum balance.", $args);
      }
      else {
        $message = t("This transaction would put account '@acc' on node @node @diff below its maximum balance.", $args);
      }
      throw new \Exception(Markup::create(strip_tags($message)));
    }
    // Convert remote addresses to local wallets.
    foreach($additional as &$entry) {
      $entry->payee = Functions::convertCCaddressToWalletId($entry->payee, $entry->metadata);
      $entry->payer = Functions::convertCCaddressToWalletId($entry->payer, $entry->metadata);
      $entry = (array)$entry;
      $entry['metadata'] = (array)$entry['metadata'];
    }
    return $additional;
  }

  public function transactionChangeState(string $uuid, string $target_state) : void {
    try {
      parent::transactionChangeState($uuid, $target_state);
    }
    catch(CCError $e) {
      throw new \Exception($e->getMessage());
    }
  }

  /**
   * Log every request to the network.
   * @param string $endpoint
   * @return \stdClass|NULL
   */
  protected function request(string $endpoint = '/') :\stdClass|NULL {
    $headers = $this->options[RequestOptions::HEADERS] ? array_map( // cumbersome...
      function ($v, $k){return "$k: $v";},
      $this->options[RequestOptions::HEADERS],
      array_keys($this->options[RequestOptions::HEADERS])
    ) : [];
    try {
      $data = parent::request($endpoint);
    }
    catch(CCFailure $e) {
      // Logger doesn't cope with the long $e->trace, so format a friendler message.
      foreach ((array)$e as $prop => $val) {
        if ($prop == 'trace' or $prop == 'translated' or substr($prop, 1, 5) == 'Error')continue;
        $body[] = "$prop : $val";
      }
      $this->logger->error(
        "<p>@headers<br />@method @domain/@endpoint<br />@request_body</p><p>@response_code: @response_message<br /><pre>@response_body</pre></p>",
        [
          '@headers'  => Markup::create(implode('<br />', $headers)),
          '@method'   => strtoupper($this->method),
          '@domain'   => $this->baseUrl,
          '@endpoint' => $endpoint,
          '@request_body'     => $this->options[RequestOptions::BODY] ?? "",
          '@response_code'     => $e->getCode(),
          '@response_message'  => $e->makeMessage(),
          '@response_body' => implode('<br />', $body)
        ]
      );
      // Don't give any details to the user.
      // Note this appears in the log twice, but I would want it not at all.
      throw new \Exception('There was a problem on the remote server. See the log for more details.');
    }

    $this->logger->notice(
       "<p>@headers<br />@method @domain/@endpoint<br />@body</p><p><pre>@response</pre></p>",
      [
        '@headers'  => Markup::create(implode('<br />', $headers)),
        '@method'   => strtoupper($this->method),
        '@domain'   => $this->baseUrl,
        '@endpoint' => $endpoint,
        '@body'     => $this->options[RequestOptions::BODY] ?? "",
        '@response' => print_r($data, 1)
      ]
    );
    return $data;
  }

}
