<?php

namespace Drupal\mcapi_forms;

use Drupal\mcapi\Functions;
use Drupal\mcapi_forms\Entity\McForm;
use Drupal\mcapi\Form\TransactionForm;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Plugin\Field\FieldWidget\EntryWidget;
use Drupal\Core\Render\Markup;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Entity\EntityConstraintViolationListInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Entity\EntityFieldManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form builder for transaction formed designed via the UI.
 */
class DesignedTransactionForm extends TransactionForm {

  /**
   * @var type
   */
  protected $config;

  /**
   * @var Renderer
   */
  protected $renderer;

  /**
   * @var EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * @var EntityFieldManager
   */
  protected $entityFieldManager;

  /**
   * @param type $entity_repository
   * @param type $entity_type_bundle_info
   * @param type $time
   * @param type $tempstoreFactory
   * @param type $current_request
   * @param Renderer $renderer
   * @param EntityFieldManager $entity_field_manager
   * @param EntityTypeManager $entity_type_manager
   */
  public function __construct($entity_repository, $entity_type_bundle_info, $time, $tempstoreFactory, $current_request, Renderer $renderer, EntityFieldManager $entity_field_manager, EntityTypeManager $entity_type_manager) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time, $tempstoreFactory, $current_request);
    $this->renderer = $renderer;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('tempstore.private'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('renderer'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\Core\Entity\Entity\EntityFormDisplay $form_display */
    $form_display = $form_state->get('form_display');
    $added_fields = array_filter(
      $this->entityFieldManager->getFieldDefinitions('mc_transaction', 'mc_transaction'),
      function ($f) {return !$f instanceOf \Drupal\Core\Field\BaseFieldDefinition;}
    );

    $form['#title'] = $this->config->label();
    $form['#twig_template'] = str_replace(["\r\n","\n","\r"], "<br/>", $this->config->getTwigTemplate());
    // avoid calling parent::form and adding the fields.
    $form['#process'][] = '::processForm';
    $form['#entity_builders']['update_form_langcode'] = '::updateFormLangcode';
    $form['#parents'] = [];
    $this->renderer->addCacheableDependency($form, $form_display);
    $raw_elements = EntryWidget::rawElements($this->entity->entries, 0);
    // A cumbersome way to find the non-base fields (No base fields can appear on Designed forms.)
    preg_match_all('/\{\{ *(.*?) *\}\}/', $form['#twig_template'], $matches);
    foreach ($matches[1] as $fieldname) {
      $form['#tokens'][$fieldname] = $fieldname;
      // Treat API fields and entries fields seperately
      // API field...
      if ($field_definition = @$added_fields[$fieldname]) {
        $items = $this->getEntity()->get($fieldname);
        // Make a hidden field with the value, and show the rendered value as a suffix.
        if ($preset = $this->config->getPreset($fieldname)) {
           $form[$fieldname] = [
            '#type' => 'hidden',
            '#value' => $preset
          ];
          $items->setValue($preset);
          $field_view = $this->entityTypeManager->getViewBuilder('mc_transaction')
            ->viewField($items, ['label' => 'hidden']);
          $form[$fieldname]['#suffix'] = $this->renderer->render($field_view[0]);
        }
        else { // If there is no preset value, get the widget to render it.
          $element = [
            '#title_display' => 'hidden',
            '#required' => $field_definition->isRequired()
          ];
          // Build the form for the field.
          $form[$fieldname] = \Drupal::service('plugin.manager.field.widget')
            ->getInstance(['field_definition' => $field_definition])
            ->formElement($items, 0, $element, $form, $form_state);
          $form_field_widget = &$form[$fieldname]['widget'];
          // Remove the container i.e. the field label.
          if (isset($form_field_widget['0']['value'])) {
            $form_field_widget = $form_field_widget['0']['value'];
          }
          if (isset($form_field_widget['0']['target_id'])) {
            $form_field_widget = $form_field_widget['0']['target_id'];
          }
          elseif (!empty($form_field_widget)) {
            $form_field_widget = $form_field_widget[0];
          }
          else {
            $form_field_widget['#access'] = FALSE;
          }
        }
      }
      else { // For the 4 "entries" subfields
        if ($preset = $this->config->getPreset($fieldname)) {
          switch ($fieldname) {
            case 'payee':
            case 'payer':
              $markup = Wallet::load($preset)->label();
              break;
            case 'description':
              $markup = Markup::Create($preset);
              break;
            case 'quant':
              $markup = Functions::formatCurrency($preset);
              break;
          }
          // Make a hidden field with the value, and show the rendered value as a suffix.
          $element = [
            '#type' => 'hidden',
            '#value' => $preset,
            '#suffix' => ' '.$markup.' '
          ];
        }
        else {// With no preset value, use the formElements given in EntriesItem
          switch ($fieldname) {
            case 'payee':
            case 'payer':
              $element = $raw_elements[$fieldname.'_wid'];
              break;
            case 'description':
            case 'quant':
              $element = $raw_elements[$fieldname];
              break;
          }
          // this wrapper is needed for css to highlight the errors.
          $element['#prefix'] = '<span class="form-item">';
          $element['#suffix'] = '</span>';
        }
        $form[$fieldname] = $element;
      }
    }
    $form['#twig_template'] = str_replace(["\r\n","\n","\r"], "<br/>", $this->config->getTwigTemplate());
    $form['#attached']['library'][] = 'mcapi_forms/display';
    $form['#after_build'][] = '::prepareTemplateForm';
    // Links are not being detected in the yml file for some reason.
    $form['#contextual_links']['mcapi_form'] = ['route_parameters' => []];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Send');
    // @todo check this is working.
    $form['#cache']['contexts'][] = 'user';
    return $actions;
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'designed_transaction_form';
  }


  /**
   * Provide an ID for the transaction form consistent with the keys of hook_transction_form_list().
   */
  function transactionFormId() {
    return 'designed_'.$this->config->id();
  }

  /**
   * {@inheritDoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $form_state->setValue('entries', [
      'description' => $form_state->getValue('description'),
      'quant' => $form_state->getValue('quant'),
      'payee_wid' => $form_state->getValue('payee'),
      'payer_wid' => $form_state->getValue('payer'),
      'primary' => TRUE
    ]);
    return parent::buildEntity($form, $form_state);
  }

  /**
   * Form #afterbuild callback
   * Break up the template into render elements alternating between markup and form elements.
   */
  function prepareTemplateForm($form, $form_state) {
    // Move the form elements to be placed from tokens out of the form and into
    // $replacements, stripping off the containers.
    foreach ($form['#tokens'] as $fname => $tok) {
      $replacements[$tok] = $form[$fname];
      unset($form[$fname]);
    }
    $w = 0;
    // break up the template into alternating tokens and html snippents
    $tok = strtok(' '.$form['#twig_template'], '{');// ensure the first is text
    while ($tok !== FALSE) {
      $parts[] = $tok;
      $tok = strtok('{}');
    }
    foreach ($parts as $key => $snippet) {
      $snippet = trim($snippet);
      if ($key % 2 == 0) { //even numbers are text
        $form['snippet-'.$key] = [
          '#markup' => $snippet,
          '#weight' => $w++,
          '#parents' => []
        ];
      }
      // Replace the fields
      elseif (isset($replacements[$snippet])) { // odd numbers are tokens
        $element = $replacements[$snippet];
        $element['#weight'] = $w++;
        if (empty($element['#parents'])) {
          $element['#parents'] = $element['#array_parents'] = [$element['#name']];
        }
        $element['#array_parents']= $element['#parents'];
        unset($element['#theme_wrappers']);
        $form[$snippet] = $element;
      }
      else {
        throw new \Exception("Unrecognised token in designed transaction form: $snippet");
      }
    }
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function getEntityFromRouteMatch(RouteMatchInterface $route_match, $entity_type_id) {
    $route_name = $route_match->getRouteName();
    $mc_form_id = substr($route_name, strrpos($route_name, '.') + 1);
    $this->config = McForm::load($mc_form_id);
    $this->request->attributes->set('mc_workflow', $this->config->getWorkflowId());
    return parent::getEntityFromRouteMatch($route_match, $entity_type_id);
  }

  /**
   * Ensure that errors to the payer, payee, quant and description fields are processed.
   */
  protected function getEditedFieldNames(FormStateInterface $form_state) {
    $fields = parent::getEditedFieldNames($form_state);
    $fields[] = 'entries';
    return $fields;
  }

  /**
   * {@inheritDoc}
   */
  protected function flagViolations(EntityConstraintViolationListInterface $violations, array $form, FormStateInterface $form_state) {
    foreach ($violations as $offset => $violation) {
      $propPath = $violation->getPropertyPath();
      $fieldname = substr($propPath, strrpos($propPath, '.') + 1);
      switch ($fieldname) {
        // Drupals' form.css expects these fields to be in a <div class="form-item"> to show the error message
        case 'payee_wid':
          $element = $form['payee'];
          break;
        case 'payer_wid':
          $element = $form['payer'];
          break;
        case 'quant':
        case 'description':
          $element = $form[$fieldname];
          break;
        default:
          break 2;
      }
      $form_state->setError($element, $violation->getMessage());
      $violations->remove($offset);
    }
    parent::flagViolations($violations, $form, $form_state);

  }
}
