<?php

namespace Drupal\mcapi\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\mcapi\Functions;
use Drupal\mcapi\Entity\Wallet;

/**
 * Builds a list of transaction forms.
 */
class Metrics extends ControllerBase {

  /**
   * Build a page showing all the trading stats
   */
  public function buildPage() {
    $transactionStorage = \Drupal::entityTypeManager()->getStorage('mc_transaction');
    $currency = \Drupal::config('mcapi.settings')->get('currency');
    $build = [
      '#title' => $this->t('@currency metrics', ['@currency' => $currency['name']])
    ];
    $form_state = new \Drupal\Core\Form\FormState;
    if ($filterform = \Drupal::formBuilder()->buildForm('Drupal\mcapi\Form\TransactionStatsFilterForm', $form_state)) {
      $build['filter'] = $filterform;
    }
    $build['#cache_contexts'] = ['accounting'];

    $conditions = [];
    if (isset($_SESSION['transaction_stats_filter'])) {
      $conditions = array_filter($_SESSION['transaction_stats_filter']);
    }
    elseif ($earliest = $transactionStorage->firstCreated()) {
      // $since needs to be set to the second previous to the first transaction;
      $conditions['since'] = $earliest - 1;
    }

    $balances_asc = $trades = $volumes = $incomes = $spending = [];

    $rows = $transactionStorage->ledgerStateByWallet($conditions??[]);
    foreach ($rows as $row) {
      $balances_asc[$row->wid] = (int)$row->balance;
      $trades[$row->wid] = $row->trades;
      $volumes[$row->wid] = (int)$row->income + (int)$row->expenditure;
      $incomes[$row->wid] = (int)$row->income;
      $spending[$row->wid] = (int)$row->expenditure;
    }

    asort($balances_asc);
    // Get the bottom wallets.
    for ($i = 0; $i < min(10, count($balances_asc)/2); $i++) {
      $wid = key($balances_asc);
      $wallet = Wallet::load($wid);
      if (!$wallet) {
        throw new \Exception("No wallet $wid in \Drupal\mcapi\Controller\Metrics::buildPage");
      }
      $quant = current($balances_asc);
      $bottoms[] = [
        'raw' => abs($quant),
        'link' => $wallet->toUrl(),
        'quant' => Functions::formatCurrency($quant, FALSE),
        'name' => $wallet->label(),
      ];
      next($balances_asc);
    }
    // Get the top wallets.
    $balances_desc = array_reverse($balances_asc, TRUE);
    for ($i = 0; $i < min(10, count($balances_asc)/2); $i++) {
      $wid = key($balances_desc);
      $wallet = Wallet::load($wid);
      if (!$wallet) {
        throw new \Exception("No wallet $wid in \Drupal\mcapi\Controller\Metrics::buildPage");
      }
      $quant = current($balances_desc);
      $tops[] = [
        'raw' => $quant,
        'link' => $wallet->toUrl(),
        'quant' => Functions::formatCurrency($quant, FALSE),
        'name' => $wallet->label(),
      ];
      next($balances_desc);
    }

    if (!empty($balances_desc)) {
      $build['balances_chart'] = [
        '#theme' => 'mc_extreme_balances',
        '#largest' => max(max($balances_asc), abs(min($balances_asc))),
        '#tops' => array_filter($tops, function($t){return $t['raw'] > 0;}),
        '#bottoms' => array_filter($bottoms, function($t){return $t['raw'] > 0;}),
        '#weight' => -1,
        '#class' => ['extreme-balances'],
      ];
    }
    $build['trades_per_user_wallet'] = [
      '#title' => t('Trades per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'trades_per_user_wallet',
      '#users_only' => TRUE,
      '#format_vals' => FALSE,
      '#data' => $trades,
      '#top' => 5,
    ];
    $build['volume_per_user_wallet'] = [
      '#title' => t('Volume per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'volume_per_user_wallet',
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $volumes,
      '#top' => 5,
    ];
    $build['income_per_user_wallet'] = [
      '#title' => t('Income per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'income_per_user_wallet',
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $incomes,
      '#top' => 5,
    ];
    $build['spending_per_user_wallet'] = [
      '#title' => t('Expenditure per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'spending_per_user_wallet',
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $spending,
      '#top' => 5,
    ];

    //@todo support time periods
    list(, , $period, $period_name) = $this->periodQueryParams($conditions);
    list($dates, $trades, $volumes, $wallets) = $transactionStorage->historyPeriodic($period, $conditions);
    $params = [
      '@span' => strtolower($period_name), // this should be already translated, somehow.
    ];
    $build['trades_history'] = [
      '#theme' => 'mc_timeline',
      '#title' => t('Numbers of trades per @span', $params),
      '#points' => array_combine($dates, $trades),
      '#width' => 800,
      '#height' => 200,
      '#format_curr' => FALSE
    ];
    $build['participation_history'] = [
      '#theme' => 'mc_timeline',
      '#title' => t('Numbers of traders per @span', $params),
      '#points' => array_combine($dates, $wallets),
      '#width' => 800,
      '#height' => 200,
      '#format_curr' => FALSE
    ];
    $build['volume_history'] = [
      '#theme' => 'mc_timeline',
      '#title' => t('Volumes traded per @span', $params),
      '#points' => array_combine($dates, $volumes),
      '#width' => 800,
      '#height' => 200,
      '#format_curr' => TRUE
    ];
    $build['#attached']['library'][] = 'mcapi/currency';
    return $build;
  }

  /**
   * Work out the start, end and frequency to poll the db.
   *
   * Less than a week, invalid.
   * Less than 100 days.
   * Less than 100 weeks, weekly.
   * Less than 100 months, monthly.
   * Otherwise annual.
   *
   * @param array $conditions
   * @return array
   *   $from, $to, the time period, and the timeperiod translated
   */
  public function periodQueryParams(array $conditions) : array {
    $since = isset($conditions['since']) ? $conditions['since'] : 0;
    $until = isset($conditions['until']) ? $conditions['until'] : \Drupal::time()->getRequestTime();
    $span = $until - $since;
    $day = 86400;
    // $year = 31560192;
    if ($span < $day * 7) {
      // This won't produce very good results.
      $from = date('d-m-Y', $since);
      $to = date('d-m-Y', $until + $day);
      $groupby_translated = t('Day');
      $groupby = 'day';
    }
    // 100 days.
    elseif ($span < $day * 100) {
      $from = date('d-m-Y', $since);
      $to = date('d-m-Y', $until + $day);
      $groupby_translated = t('Day');
      $groupby = 'day';
    }
    // 100 weeks.
    elseif ($span < $day * 700) {
      // Get the beginning of the week.
      $from = date('Y', $since) . 'W' . date('W', $since);
      // The beginning of the week after the end.
      $to = date('Y', $since) . 'W' . (date('W', $since) + 1);
      $groupby_translated = t('Week');
      $groupby = 'week';
    }
    // 100 months.
    elseif ($span < $day * 3000) {
      $from = date('Y', $since);
      // The beginning of the week after the end.
      $to = date('Y', $since) + 1;
      $groupby_translated = t('Month');
      $groupby = 'month';
    }
    else {
      $from  = date('Y', $since);
      $to = date('Y', $until) + 1;
      $groupby_translated = t('Year');
      $groupby = 'year';
    }
    return [strtotime($from), strtotime($to), $groupby, $groupby_translated];
  }

}
