<?php

namespace Drupal\mcapi\Element;

use Drupal\mcapi\Functions;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\FormElement;

/**
 * Provides form element to display amounts
 *
 * @FormElement("worth")
 */
class Worth extends FormElement {

  /**
   * {@inheritDoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#input' => TRUE,
      '#value_callback' => [
        [$class, 'valueCallback'],
      ],
      '#element_validate' => [
        [$class, 'validate'],
      ],
      '#process' => [
        [$class, 'process'],
      ],
      '#attached' => ['library' => ['mcapi/worth.element']],
      '#theme_wrappers' => ['form_element'],
      '#tree' => TRUE
    ];
  }

  /**
   * Form element processor callback for 'worth'.
   */
  public static function process($element) {
    // Render the widget for a known currency
    $value = $element['#default_value'] ?? 0;
    $format = \Drupal::config('mcapi.settings')->get('currency')['format'];
    $value_parts = Functions::formatCurrencyParts((int)$value);
    $placeholder_parts = Functions::formatCurrencyParts((int)@$element['#placeholder']);
    $element['main'] = static::formatMain($format['main'], $value_parts['main'], $placeholder_parts['main']);
    $element['main']['#field_prefix'] = $format['f1'];
    $element['main']['#field_suffix'] = $format['f2'];
    if (isset($format['subdivision'])) {
      $element['subdivision'] = static::formatSubdivision($format['subdivision'], (int)$value_parts['subdivision'], $placeholder_parts['subdivision']);
      if (isset($format['f3'])) {
        $element['subdivision']['#field_suffix'] = $format['f3'];
      }
    }
    $parts['main'] = $element['main']['#value'] ?? $element['main']['#default_value'];
    $parts['subdivision'] = $element['subdivision']['#default_value'] ?? 0;
    $element['#value'] = static::unformat($parts);
    // This is a hack that prevents FormValidator::dovalidateForm from,
    // registering the field as empty when the two subfields have been
    // calculated. Validation still takes place normally.
    unset($element['#needs_validation']);
    //$element['#attributes']['class'] = ['worth-container'];
    return $element;
  }
  /**
   * Generate the form element for the main currency units.
   *
   * @param string $format
   * @param int $num
   * @return array
   */
  private static function formatMain(string $format, int $num, int $placeholder) {
    $dotpos = strrpos($format, '.');
    if ($dotpos and is_int($num)) {
      $num .= str_pad('.', $dotpos+1, '0', STR_PAD_RIGHT);
      $placeholder .= str_pad('.', $dotpos+1, '0', STR_PAD_RIGHT);
    }
    $element = [
      '#type' => 'textfield',
      '#default_value' => $num,
      '#size' => 6,
      '#weight' => 1,
      '#pattern' => $dotpos ? "[0-9]+(\.[0-9]{{$dotpos}})?" : '[0-9]+',
      '#placeholder' => $placeholder
      #'#name' => "entries[quant][main]"
    ];
    if ($dotpos) {
      $element['#step'] = pow(10, -$dotpos);
    }
    return $element;
  }

  /**
   * Generate the form element for the subdivision currency units.
   *
   * @param string $format
   * @param int $num
   * @return array
   */
  private static function formatSubdivision(string $format, int $num, $placeholder) {
    // If there is a slash in the format, then prepare a dropdown
    if (strpos($format, '/')) {
      $options = [];
      list($format, $divisor) = explode('/', $format);
      $base = $format + 1;
      $step = $base / $divisor;
      for ($j = 0; $j < $divisor; $j++) {
        $v = $j * $step;
        $options[intval($v)] = $v;
      }
      // If a preset value isn't in the $options ignore the options and use
      // the numeric sub-widget.
      if (array_key_exists($num, $options)) {
        $element = [
          '#type' => 'select',
          '#options' => $options,
          '#weight' => 3,
          '#placeholder' => $placeholder
        ];
      }
      else {
        $size = strlen($format);
        $element = [
          '#type' => 'number',
          '#min' => 0,
          '#step' => $step,
          '#size' => $size,
          '#maxlength' => $size,
          '#attributes' => ['style' => 'width:' . ($size + 3) . 'em;'],
          '#max' => $format, //Which is already all the nines
          '#placeholder' => $placeholder
        ];
        // Leave the main value field empty & put zero placeholders in other
        // fields.
        $element['#placeholder'] = str_pad('', $size, '0');
      }
    }
    else {// Just enter the number of sub units
      $element = [
        '#type' => 'number',
        '#default_value' => $num,
        '#min' => 0,
        '#step' => 1,
        // #size has no effect in opera.
        '#size' => strlen($format)+2, // Spinner takes about 2 chars
        '#maxlength' => strlen($format),
        '#max' => $format, // E.G. 59 minutes
        '#weight' => 3
      ];
    }
    $element['#attributes']['class'] = ['worth-subdivision'];
    return $element += [
      '#default_value' => $num,
      '#theme_wrappers' => [],
      '#weight' => 3
    ];
  }

  /**
   * {@inheritDoc}
   * This is called before te two sub elements
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if (is_array($input)) {
      return static::unformat($input);
    }
  }

  /**
   * convert the field contents back to a raw number.
   * @param array $parts
   * @return int
   */
  private static function unformat(array $parts) : int {
    $parts['main'] = $parts['main'] ? $parts['main'] + 0 : 0; //Make this numeric if it is null.
    $format = \Drupal::config('mcapi.settings')->get('currency')['format'];
    if (empty($format['subdivision'])) {
      $raw_num = $parts['main'];
      if (strpos($format['main'], '.')) {
        $raw_num *= pow(10, strrpos($format['main'], '.'));
      }
    }
    else {
      list($atoms) = explode('/',  $format['subdivision']);
      if (!is_numeric($atoms)) {// could happen in migration. @todo in premigration check with a regex.
        trigger_error("problem converting currency format");
      }
      $divisor = (int)$atoms + 1;
      $fractions = is_numeric($parts['subdivision']) ? $parts['subdivision'] : 0;
      $raw_num = $parts['main'] * $divisor + $fractions;
    }
    return intval($raw_num);
  }

  /**
   * Validate callback.
   *
   * @note the value is set here, coz valuecallback doesn't work when there are subfields.
   */
  public static function validate(&$element, FormStateInterface $form_state) {
    $value_parts = [
      'main' => $element['main']['#value'],
      'subdivision' => $element['subdivision']['#value']??0
    ];
    $val = $element['#value'] = static::unformat($value_parts);

    $form_state->setValueForElement($element, $val);
    if ($val) {
      unset($element['#required_but_empty']);
    }
    // Check for allowed zero values
    if ($val < 0) { // Should never happen!
      $form_state->setError($element, t('Negative amounts not allowed: %val'));
    }
    if (isset($element['#min']) and $val < $element['#min']) {
      $form_state->setError($element['main'], t('Min value is %val', ['%val' => Functions::formatCurrency($element['#min'], FALSE)]));
    }
    if (isset($element['#max']) and $val > $element['#max']) {
      $form_state->setError($element['main'], t('Max value is: %val', ['%val' => Functions::formatCurrency($element['#max'], FALSE)]));
    }
  }

}
