<?php

namespace Drupal\mcapi\Entity\Access;

use Drupal\Core\Entity\EntityAccessControlHandler;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityHandlerInterface;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Query\WalletQuery;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Messenger\Messenger;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines an access controller option for the mc_transaction entity.
 */
class TransactionAccessControlHandler extends EntityAccessControlHandler implements EntityHandlerInterface {

  /**
   * @var Drupal\mcapi\Entity\Query\WalletQuery
   */
  protected $walletQuery;

  /**
   * @var Drupal\Core\Messenger\Messenger
   */
  protected $messenger;

  /**
   *
   * @param EntityTypeInterface $entity_type
   * @param WalletQuery $wallet_query
   * @param Messenger $messenger
   */
  public function __construct(EntityTypeInterface $entity_type, WalletQuery $wallet_query, Messenger $messenger) {
    parent::__construct($entity_type);
    $this->walletQuery = $wallet_query;
    $this->messenger = $messenger;
  }

  /**
   *
   * @param ContainerInterface $container
   * @param EntityTypeInterface $entity_type
   * @return \static
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    return new static(
      $entity_type,
      $container->get('entity_type.manager')->getStorage('mc_wallet')->getQuery(),
      $container->get('messenger')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function checkCreateAccess(AccountInterface $account, array $context, $entity_bundle = NULL) {
    // Friendly reminder that there must be at least 2 wallets on the system.
    if ($this->walletQuery->accessCheck(TRUE)->count()->execute() < 2) {
      // Would be great not to have to bother with this check.
      $this->messenger->addWarning('Transaction forms cannot be accessed until at least 2 wallets exist.');
      $result = AccessResult::forbidden('At least 2 wallets needed to transact.');
    }
    elseif ($account->hasPermission('create 3rdparty transaction')) {
      $result = AccessResult::allowed();
    }
    elseif (\Drupal::entityTypeManager()->getStorage('mc_wallet')->walletsOf($account->id())) {
      $result = AccessResult::allowed();
    }
    else {
      $result = AccessResult::forbidden('User has no wallet.');
    }
    return $result->cachePerUser();
  }

  /**
   * {@inheritDoc}
   *
   * @note There are only 2 operations, view and transition.
   */
  public function access(EntityInterface $transaction, $operation, AccountInterface $account = NULL, $return_as_object = FALSE) {
    $account = $this->prepareUser($account);
    $result = AccessResult::forbidden()->cachePerUser();
    if ($operation == 'view label' or $operation == 'view') {
      $operation = 'view';
      $cid = 'mc_transaction:' . $transaction->id();
      if (($return = $this->getCache($cid, $operation, 'und', $account)) !== NULL) {
        // Cache hit, no work necessary.
        return $return_as_object ? $return : $return->isAllowed();
      }
      if ($account->hasPermission('view all transactions') or $account->hasPermission('manage mcapi')) {
        $result = AccessResult::allowed()->cachePerUser();
      }
      else {
        // You can view this transaction if you own any of the wallets in it.
        foreach (array_chunk($transaction->allWalletIds(), 50) as $wids) {
          foreach (Wallet::loadMultiple($wids) as $wallet) {
            if ($wallet->getOwnerId() == $account->id()) {
              $result = AccessResult::allowed()->cachePerUser();
              break 2;
            }
          }
        }
      }
    }
    elseif ($operation == 'transition') { // Check the workflow.
      if ($account->hasPermission('manage mcapi')) {
        $result = AccessResult::allowed()->cachePerUser();
      }
      else {
        $roles = $transaction->workflow->entity->states[$transaction->state->value][\Drupal::request()->attributes->get('dest_state')]['actors'];
        if ($transaction->payee->getOwner() == $account->id) {
          if (in_array('payee', $roles)) {
            $result = AccessResult::allowed();
          }
        }
        elseif ($transaction->payer->getOwner() == $account->id) {
          if (in_array('payer', $roles)) {
            $result = AccessResult::allowed();
          }
        }
      }
    }
    $others = $this->moduleHandler()->invokeAll('mc_transaction_access', [$transaction, $operation, $account]);
    $return = $this->processAccessHookResults(array_merge($others, [$result]));

    if ($operation == 'view') {
      $result = $this->setCache($return, $cid, 'view', 'und', $account);
    }
    return $return_as_object ? $result : $result->isAllowed();
  }

}
