<?php

namespace Drupal\mcapi\Entity;

use Drupal\mcapi\Plugin\Field\FieldType\EntryItem;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Functions;
use Drupal\mcapi\Event\McapiEvents;
use Drupal\mcapi\Event\TransactionSaveEvent;
use Drupal\user\EntityOwnerTrait;
use Drupal\user\EntityOwnerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityConstraintViolationListInterface;

/**
 * Defines the Transaction entity.
 * @ContentEntityType(
 *   id = "mc_transaction",
 *   label = @Translation("Transaction"),
 *   label_collection = @Translation("Transactions"),
 *   handlers = {
 *     "storage" = "Drupal\mcapi\Entity\Storage\TransactionStorage",
 *     "storage_schema" = "Drupal\mcapi\Entity\Storage\TransactionStorageSchema",
 *     "view_builder" = "Drupal\mcapi\Entity\ViewBuilder\TransactionViewBuilder",
 *     "list_builder" = "\Drupal\mcapi\Entity\ListBuilder\TransactionListBuilder",
 *     "access" = "Drupal\mcapi\Entity\Access\TransactionAccessControlHandler",
 *     "form" = {
 *       "default" = "Drupal\mcapi\Form\TransactionForm",
 *       "transition" = "Drupal\mcapi\Form\TransactionConfirmForm"
 *     },
 *     "views_data" = "Drupal\mcapi\Entity\Views\TransactionViewsData",
 *     "route_provider" = {
 *       "html" = "Drupal\Core\Entity\Routing\DefaultHtmlRouteProvider",
 *     },
 *   },
 *   admin_permission = "configure mcapi",
 *   base_table = "mc_transaction",
 *   revision_table = "mc_transaction_revision",
 *   show_revision_ui = FALSE,
 *   entity_keys = {
 *     "id" = "xid",
 *     "revision" = "vxid",
 *     "uuid" = "uuid",
 *     "name" = "description",
 *     "owner" = "creator"
 *   },
 *   field_ui_base_route = "entity.mc_transaction.collection",
 *   translatable = FALSE,
 *   links = {
 *     "canonical" = "/transaction/{mc_transaction}",
 *     "collection" = "/admin/accounting/transactions"
 *   },
 * )
 */
class Transaction extends ContentEntityBase implements TransactionInterface, EntityOwnerInterface {

  use EntityChangedTrait;
  use EntityOwnerTrait;
  use StringTranslationTrait;

  /**
   * Constructor.
   *
   * @note There seems to be no way to inject anything here.
   *
   * @see SqlContentEntityStorage::mapFromStorageRecords
   */
  public function __construct(array $values, $entity_type, $bundle = FALSE, $translations = array()) {
    parent::__construct($values, $entity_type, $bundle, $translations);
  }

  /**
   * {@inheritDoc}
   */
  public function label($langcode = NULL) {
    return $this->t("Transaction #@id", ['@id' => $this->id()]);
  }

  /**
   * @inheritDoc
   *
   * The assemble phase must happen after Entity ref fields are validated
   * because entity reference fields only determine their final value during
   * validation. The assemble phase must happen before entity level validation.
   * Therefore this function validates the entity, assembles it, and validates
   * it again.
   */
  public function validate() : EntityConstraintViolationListInterface {
    $violations = parent::validate();
    /** @var EntityConstraintViolationListInterface $violations */
    if (!$violations->count()) {
      // Credcom module might generate exceptions during assembly
      try {
        // This also puts the transaction into its first defined 'state'
        Functions::getAssembledTransaction($this);
        $violations = parent::validate();
      }
      catch(\Exception $e) {
        $violation = new \Symfony\Component\Validator\ConstraintViolation(
          message: $e->getMessage(),
          messageTemplate: 'message template',
          parameters: [],
          propertyPath: 'payer_wid',
          invalidValue: $this->entries[0]->getValue(),
          root: $this->entries[0]->getValue(),
          code: $e->getCode(),
        );
        $violations->add($violation);
      }
    }
    // Post validation - add signatures and set state to validated.
    if ($violations->count() == 0) {
      if ($this->workflow->entity->creation['state'] == 'pending') {
        // Add signatures. can only happen when we are sure payer and payee fields are populated and valid
        $this->signatures->appendItem(['uid' => $this->payee->getOwnerId()]);
        $this->signatures->appendItem(['uid' => $this->payer->getOwnerId()]);
        try {
          $this->signatures->sign(\Drupal::currentUser()->id());
        }
        catch(\Exception $e) {} // 3rd party transactions wouldn't need the current users's signature
      }
      $this->state->value = 'validated';
    }
    return $violations;
  }

  /**
   * {@inheritDoc}
   * //need to handle mass transactions here because there's no appropriate hook
   */
  public static function preCreate(EntityStorageInterface $storage, array &$values) {
    $values += [
      // If creator is 0 that means drush or credcom must have created it.
      'creator' => static::getDefaultEntityOwner(), //currentuser
      'state' => 'temp',
      'payee' => [],
      'payer' => [],
    ];

    $first = TRUE;
    if (isset($values['entries'])) {
      foreach ($values['entries'] as &$e) {
        $e['primary'] = $first;
        $first = FALSE;
      }
    }
    else {
      // Multiple payers or payees indicates a mass transaction.
      foreach ((array)$values['payee'] as $payee_wid) {
        foreach ((array)$values['payer'] as $payer_wid) {
          $values['entries'][] = [
            'payer_wid' => $payer_wid,
            'payee_wid' => $payee_wid,
            'description' => $values['description'],
            'quant' => $values['quant'],
            'primary' => $first
          ];
          $first = FALSE;
        }
      }
    }
    unset($values['payee'], $values['payer']);
  }

  /**
   * {@inheritDoc}
   */
  public function getCacheTagsToInvalidate() {
    $tags = parent::getCacheTagsToInvalidate();
    // Invalidate tags of wallets
    foreach (array_chunk($this->allWalletIds(), 50) as $wids) {
      foreach (Wallet::loadMultiple($wids) as $wallet) {
        $tags = array_merge_recursive($tags, $wallet->getCacheTagsToInvalidate());
      }
    }
    return $tags;
  }

  public function getCacheContexts() {
    $contexts = parent::getCacheContexts();
    // Add the user context because links are shown per user.
    $contexts[] = 'user:'.\Drupal::currentUser()->id();
    $contexts[] = 'user';
    return $contexts;
  }

  /**
   * {@inheritDoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type)
      + static::ownerBaseFieldDefinitions($entity_type);

    $fields['entries'] = BaseFieldDefinition::create('entry')
      ->setLabel(t('Entries'))
      ->setDescription(t('One or more flows of currency between accounts.'))
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED)
      ->setRequired(TRUE)
      ->addConstraint('DifferentWallets')
      ->setRevisionable(FALSE)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', ['weight' => 4]);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Date'))
      ->setDescription(t('The time that the transaction was created.'))
      ->setRevisionable(FALSE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'timestamp',
        'weight' => 0,
      ])
      ->setDisplayOptions('form', [
        'type' => 'datetime_timestamp',
        'weight' => 10,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['workflow'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Workflow'))
      ->setDescription(t('The workflow path for the transaction'))
      ->setSetting('target_type', 'mc_workflow')
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayOptions('form', ['region' => 'hidden'])
      ->setRevisionable(FALSE)
      ->setReadOnly(TRUE)
      ->setRequired(TRUE);

    $fields['state'] = BaseFieldDefinition::create('string')
      ->setLabel(t('State'))
      ->setDescription(t('Completed, pending, disputed, etc.'))
      ->setDefaultValue('temp')
      ->setDisplayConfigurable('view', TRUE)
      ->setRevisionable(TRUE)
      ->setDisplayOptions('form', ['region' => 'hidden'])
      ->addPropertyConstraints('value', [
        'AllowedValues' => ['callback' => ['Drupal\mcapi\Entity\Transaction', 'states']],
      ])
      ->setRequired(TRUE);

    $fields['signatures'] = BaseFieldDefinition::create('mc_signature')
      ->setLabel(t('Signatures'))
      ->setDescription(t('User IDs with dates of signing.'))
      ->setRevisionable(FALSE)
      ->setCardinality(FieldStorageDefinitionInterface::CARDINALITY_UNLIMITED)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', [
        'label' => 'hidden',
        'type' => 'signatures',
        'weight' => 2
      ]);

    return $fields;
  }

  /**
   * {@inheritDoc}
   *
   * Workaround for when the content_translation module tries to link to the
   * original language before the entity is saved and an id is available on transaction/0/save
   * I don't know why the entity definition 'translatable = FALSE' doesn't apply
   */
  public function hasLinkTemplate($rel) {
    if (!$this->isNew()) {
      return parent::hasLinkTemplate($rel);
    }
    return FALSE;
  }

  /**
   * @param array $params
   * @throws \Exception
   */
  public function addEntry(array $params) : EntryItem|NULL {
    if (!isset($params['payee_wid'])){
      $params['payee_wid'] = $params['payee'];
    }
    if (!isset($params['payer_wid'])) {
      $params['payer_wid'] = $params['payer'];
    }
    unset($params['payer'], $params['payee']); // these will be computed fields.
    if ($params['payer_wid'] == $params['payee_wid']) {
      return NULL;
    }
    $params['description'] = (string)$params['description'];
    $this->entries->appendItem($params);
    return $this->entries->offsetGet($this->entries->count()-1);
  }

  /**
   * {@inheritDoc}
   */
  function &__get($name) {
    switch ($name) {
      case 'payer':
        $val = $this->entries->payer;
        break;
      case 'payer_wid':
        $val = $this->entries->payer_wid;
        break;
      case 'payee':
        $val = $this->entries->payee;
        break;
      case 'payee_wid':
        $val = $this->entries->payee_wid;
        break;
      case 'quant':
        $val = $this->entries->quant;
        break;
      case 'description':
        $val = $this->entries->description;
        break;
      case 'version':
        $revisions = \Drupal::entityTypeManager()
          ->getStorage('mc_transaction')
          ->revisionIds($this->id());
        $val = count($revisions); // @todo Check this is the number we want.
    }
    if (isset($val)) {
      return $val;
    }
    return parent::__get($name);
  }

  /**
   * @return int[]
   */
  function allWalletIds() : array {
    $wallet_ids = [];
    foreach ($this->entries as $entry_item) {
      $wallet_ids[] = $entry_item->payee_wid;
      $wallet_ids[] = $entry_item->payer_wid;
    }
    // This method could be called on a new transaction with wallet fields not populated.
    return array_unique($wallet_ids);
  }

  /**
   * Determine if a wallet is involved in this transaction.
   *
   * @param int $wid
   * @return bool
   */
  function involves(int $wid) : bool {
    $result = FALSE;
    foreach ($this->entries as $entry_item) {
      if ($entry_item->payee_wid  == $wid or $entry_item->payer_wid == $wid) {
        $result = TRUE;
        break;
      }
    }
    return $result;
  }

  /**
   * Get a list of transaction states
   * @return array
   */
  static function states() : array {
    $states = array_keys(mc_states());
    return array_combine($states, $states);
  }

  /**
   * The time the transaction was created.
   * @return int
   */
  function getCreatedTime() : int {
    return $this->created->value;
  }

  /**
   * Utility
   */
  function newVersion(string $new_state) : static {
    if (!$this->oldState) {
      $this->oldState = $this->state->value;
    }
    $this->state->value = $new_state;
    if (!$this->isNew()) {
      $this->setNewRevision(TRUE);
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  function presave($transaction_storage) {
    $this->set('state', $this->workflow->entity->creation['state']);
    if ($this->state->value == 'pending' and !$this->signatures->getUnsigned()) {
      // All signatures present, change the state at the last minute.
      $this->state->value = 'completed';
    }
  }

  function insert() {
    \Drupal::service('event_dispatcher')->dispatch(
      new TransactionSaveEvent($this, ['old_state' => $this->oldState ?? 'temp']),
      McapiEvents::ACTION
    );
  }

  function update() {
    \Drupal::service('event_dispatcher')->dispatch(
      new TransactionSaveEvent($this, ['old_state' => $this->oldState]),
      McapiEvents::ACTION
    );
  }

}
