<?php

namespace Drupal\mcapi\EventSubscriber;

use Drupal\mcapi\Event\McapiEvents;
use Drupal\mcapi\Event\TransactionSaveEvent;
use Drupal\mcapi\Entity\Transaction;
use Drupal\user\Entity\User;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Mail\MailManager;
use Drupal\Core\Logger\LoggerChannelFactory;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;


/**
 * Subscriber showing the two Mcapi events.
 */
class TransactionSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * @var Drupal\Core\Logger\LoggerChannel
   */
  private LoggerChannel $logger;

  /**
   * @var int
   */
  private int $currentUserId;

  /**
   * @var MailManager
   */
  private MailManager $mailManager;

  /**
   * mcapi.settings
   * @var Config
   */
  protected Config $config;

  /**
   * Constructor
   *
   * @param ConfigFactoryInterface $config_factory
   * @param AccountInterface $account
   * @param MailManager $mail_manager
   * @param LoggerChannelFactory $logger_factory
   */
  function __construct(ConfigFactoryInterface $config_factory, AccountInterface $account, MailManager $mail_manager, LoggerChannelFactory $logger_factory) {
    $this->config = $config_factory->get('mcapi.settings');
    $this->currentUserId = $account->id();
    $this->mailManager = $mail_manager;
    $this->logger = $logger_factory->get('mcapi');
  }

  /**
   * {@inheritDoc}
   */
  public static function getSubscribedEvents() : array {
    return [
      McapiEvents::ACTION => ['notifyTransactees'],
    ];
  }

  /**
   * Notify transactees when a transaction changes state.
   *
   * @param TransactionSaveEvent $event
   *   $event->getSubject() gives the transaction.
   *   $event->getArguments() yields 'old_state'
   * @param string $eventName
   *   The machine name of the event.
   * @param ContainerAwareEventDispatcher $container
   *   The container.
   */
  public function notifyTransactees(TransactionSaveEvent $event, $eventName, ContainerAwareEventDispatcher $container) {
    $transaction = $event->getSubject();
    $args = $event->getArguments();
    $payee_user = $transaction->payee->getOwner();
    $payer_user = $transaction->payer->getOwner();
    if ($payee_user->id() <> $this->currentUserId) {
      $this->notify($payee_user, $transaction);
    }
    if ($payer_user->id() <> $this->currentUserId) {
      $this->notify($payer_user, $transaction);
    }
  }

  /**
   * Generate a basic transaction notification.
   * @param User $recipient
   * @param Transaction $transaction
   */
  private function notify(User $recipient, Transaction $transaction) {
    if (\Drupal::State()->get('mcapi_mail_stop')) {
      return;
    }
    if ($transaction->state->value == 'pending') {
      if (!$transaction->signatures->hasSigned($recipient->id())) {
        if ($this->config->get('signature_notification.subject') and $this->config->get('signature_notification.body')) {
          $message = $this->mailManager->mail(
            'mcapi',
            'signhere',
            $recipient->getEmail(),
            $recipient->getPreferredLangcode(),
            ['user' => $recipient, 'xaction' => $transaction]
          );
          if ($message['result']) {
            $this->logger->notice(
              'Mailed %name to sign transaction @link',
              ['%name' => $recipient->getDisplayName(), '@link' => $transaction->toLink()->toString()]
            );
          }
        }
      }
    }
    elseif ($this->config->get('mail_notification.subject') and $this->config->get('mail_notification.body')) {
      \Drupal::service('plugin.manager.mail')->mail(
        'mcapi',
        'transaction_notification',
        $recipient->getEmail(),
        $recipient->getPreferredLangcode(),
        [
          'user' => $recipient,
          'xaction' => $transaction
        ]
      );
    }
  }

}
