<?php

namespace Drupal\mcapi\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form builder for mail notification templates.
 */
class MailTemplates extends ConfigFormBase {

  /**
   * @var ModuleHandlerInterface
   */
  private $moduleHandler;

  /**
   * @var Drupal\Core\Config\Config
   */
  protected $config;

  /**
   * Constructor.
   *
   * @param ConfigFactoryInterface $configFactory
   *   The Configfactory service.
   * @param ModuleHandlerInterface $module_handler
   *   The ModuleHandler service.
   */
  public function __construct(ConfigFactoryInterface $configFactory, ModuleHandlerInterface $module_handler) {
    parent::__construct($configFactory);
    $this->config = $this->config('mcapi.settings');
    $this->moduleHandler = $module_handler;
  }

  /**
   * Injections.
   */
  static public function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $i = 0;
    $form['notify'] = [
      '#title' => $this->t('Notify transactees', [], ['context'=>'The other party to the transaction.']),
      '#description' => $this->t('Email a user when someone makes a transaction with them.'),
      '#type' => 'checkbox',
      '#default_value' => !empty($this->config->get('mail_notification.subject')) and !empty($this->config->get('mail_notification.body')),
      '#weight' => $i++,
    ];

    $token_description = \Drupal::moduleHandler()->moduleExists('token') ?
      $this->t('Tokens are available.') :
      $this->t('Enable the token module to browse available tokens.');

    $form['mail_notification'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Notification mail'),
      '#tree' => TRUE,
      'subject' => [
        '#title' => $this->t('Notification subject'),
        '#description' => $token_description,
        '#type' => 'textfield',
        '#default_value' => $this->config->get('mail_notification.subject'),
        '#weight' => 1,
        '#states' => [
          'required' => [
            [':input[name="notify"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      'body' => [
        '#title' => $this->t('Mail body'),
        '#description' => $token_description,
        '#type' => 'textarea',
        '#weight' => $i++,
        '#default_value' => $this->config->get('mail_notification.body'),
        '#weight' => 2,
        '#states' => [
          'required' => [
            [':input[name="notify"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      '#weight' => $i++,
      '#states' => [
        'visible' => [
          [':input[name="notify"]' => ['checked' => TRUE]]
        ]
      ]
    ];

    $form['notify_signatories'] = [
      '#title' => $this->t('Notify signatories', [], ['context' => 'Users who should sign a transaction.']),
      '#description' => $this->t('Email a user when a transaction requires their signature.'),
      '#type' => 'checkbox',
      '#default_value' => !empty($this->config->get('signature_notification.subject')) and !empty($this->config->get('signature_notification.body')),
      '#weight' => $i++,
    ];

    // The mail template is REQURED if any of the type->notify flags is checked.
    $form['signature_notification'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Signatures mail'),
      '#tree' => TRUE,
      'subject' => [
        '#title' => $this->t('Notification subject'),
        '#type' => 'textfield',
        '#default_value' => $this->config->get('signature_notification.subject'),
        '#weight' => 1,
        '#states' => [
          'required' => [
            [':input[name="notify_signatories"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      'body' => [
        '#title' => $this->t('Mail body'),
        '#description' => $this->t("SHOULD include the token [xaction:sign-link]"),
        '#type' => 'textarea',
        '#weight' => $i++,
        '#default_value' => $this->config->get('signature_notification.body'),
        '#weight' => 2,
        '#states' => [
          'required' => [
            [':input[name="notify_signatories"]' => ['checked' => TRUE]]
          ]
        ]
      ],
      '#states' => [
        'visible' => [
          [':input[name="notify_signatories"]' => ['checked' => TRUE]]
        ]
      ],
      '#weight' => $i++
    ];

    if (\Drupal::moduleHandler()->moduleExists('token')) {
      $form['mail_template']['tokens'] = [
        '#theme' => 'token_tree_link',
        '#token_types' => ['user', 'xaction', 'site', 'current-date'],
        '#global_types' => FALSE,
        '#weight' => $i++
      ];
    }
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config
      ->set('mail_notification', $form_state->getValue('mail_notification'))
      ->set('signature_notification', $form_state->getValue('signature_notification'))
      ->save();
    parent::submitForm($form, $form_state);
  }


  /**
   * {@inheritDoc}
   */
  protected function getEditableConfigNames() {
    return ['mcapi.settings'];
  }


  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mcapi_mail_form';
  }


}
