<?php

namespace Drupal\mcapi\Form;

use Drupal\Core\Url;
use Drupal\user\Entity\User;
use Drupal\Core\KeyValueStore\KeyValueFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityConfirmFormBase;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Render\Renderer;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A form for confirming all operations on a transaction.
 */
class TransactionConfirmForm extends EntityConfirmFormBase {

  /**
   * @var ContainerAwareEventDispatcher
   * @deprecated;
   */
  private $eventDispatcher;

  /**
   * @var Renderer
   */
  private $renderer;

  /**
   * @var Drupal\mcapi\Entity\ViewBuilder\TransactionViewBuilder
   */
  protected $transactionViewBuilder;

  /**
   * The destination state of the transaction.
   * @var string
   */
  protected $op;


  /**
   * @param EntityTypeManager $entity_type_manager
   * @param ContainerAwareEventDispatcher $event_dispatcher
   * @param Renderer $renderer
   * @param DatabaseStorage $key_value
   */
  public function __construct(EntityTypeManager $entity_type_manager, ContainerAwareEventDispatcher $event_dispatcher, Renderer $renderer, KeyValueFactoryInterface $key_value) {
    $this->transactionViewBuilder = $entity_type_manager->getViewBuilder('mc_transaction');
    $this->eventDispatcher = $event_dispatcher;
    $this->renderer = $renderer;
    $this->op = $this->getRequest()->attributes->get('dest_state');
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('event_dispatcher'),
      $container->get('renderer'),
      $container->get('keyvalue.database')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function init($form_state) {
    parent::init($form_state);
    if (!$this->entity->isNew()) {
      $this->entity->newVersion($this->op);
    }
    $this->entity->setValidationRequired(FALSE);
  }

  /**
   * {@inheritDoc}
   */
  public function buildform(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $form['#title'] = $this->getQuestion();
    $form['#attributes']['class'][] = 'transaction-operation-form';
    $certificate = $this->transactionViewBuilder->view($this->entity, 'full');
    // Remove the state-erased class coz its not erased yet.
    $key = array_search('state-erased', $certificate['#attributes']['class']);
    if ($key !== FALSE) {
      unset($certificate['#attributes']['class'][$key]);
    }
    $form['description']['#markup'] = \Drupal::service('renderer')->render($certificate);
    $form['description']['#weight'] = -1;

    // By default, render the form using theme_confirm_form().
    $form += ['#theme' => 'confirm_form'];
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Use the entity as is, don't rebuild it
    $transaction = $this->entity;
    if ($transaction and $this->op <> 'view') {
      try {
        $transaction->save();
      }
      catch(\Exception $e) {
        \Drupal::messenger()->addError(t('There was a problem: @message', ['@message' => $e->getMessage()]));
        $form_state->setRebuild();
        return;
      }
      $form_state->setRedirectUrl($transaction->toUrl());
    }
    // This assumes no non-standard workflow states.
    // @todo move the logic to each workflow state?
    // That might imply that workflow states are plugins.
    switch ($this->op) {// Destination state
      case 'deleted':
        $transaction->delete();
        $form_state->setRedirectUrl(Url::fromUserInput('/'));// redirecto to front page
        $message = $this->t('The transaction has been completely deleted');
        break;
      case 'pending':
        if ($uids = $transaction->signatures->getUnsigned()) {
          $names = array_map(function ($u){return $u->getDisplayName();}, User::loadMultiple(array_filter($uids)));
          if (in_array(0, $uids)) { // only possible with credcom module, but can't do this with a hook.
            $names[] = t('The remote party');
          }
          $message = $this->t(
            'The transaction will complete when approved by @names.',
            ['@names' => implode($names)]
          );
          break;
        }
      case 'completed':
        $message = $this->t('The transaction has been saved.');
        break;
      case 'erased':
        $message = $this->t('The transaction has been marked as erased.');
        break;
      default:
        throw new \Exception('Unknown transaction state: '.$this->op);
    }
    if (isset($message)) {
      $this->messenger()->addMessage($message);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function getCancelUrl(): Url {
    if ($this->entity->isNew()) {
      $cancel_url = new Url('user.page');
    }
    else {
      $cancel_url = $this->entity->toUrl();
    }
    return $cancel_url;
  }

  /**
   * {@inheritDoc}
   */
  public function getCancelText() {
    return $this->t('Cancel');
  }

  /**
   * {@inheritDoc}
   */
  public function getConfirmText() {
    return $this->t("I'm sure!");
  }

  /**
   * {@inheritDoc}
   */
  public function getQuestion() {
    return $this->t('Are you sure?');
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mc_transaction_' . $this->getOperation() . '_confirm_form';
  }

  /**
   * {@inheritDoc}
   */
  public function getBaseFormId() {
    return 'mc_transaction_confirm_form';
  }

  /**
   * {@inheritDoc}
   */
  protected function prepareEntity() {
    // Showing the form for the first time.
    return $this->getRouteMatch()->getParameter('mc_transaction');
  }

  /**
   * {@inheritDoc}
   */
  public function buildEntity($element, $form_state) {
    // Called on form submission
    return $this->entity;
  }

}
