<?php

namespace Drupal\mcapi\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;

/**
 * Build form to edit, workflow config entity
 */
class WorkflowEdit extends EntityForm {

  protected $allStates;

  function init(FormStateInterface $form_state) {
    parent::init($form_state);
    $states = array_keys(mc_states());
    $this->allStates = array_combine($states, $states);
    $this->allStates = $this->removeState($this->allStates, 'temp');
  }

  /**
   * {@inheritDoc}
   */
  public function getFormId() {
    return 'mcapi_workflow_form';
  }

  /**
   * {@inheritDoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $w = 0;
    $form = parent::form($form, $form_state);
    $form['#title'] = $this->entity->isNew() ?
      $this->t('New transaction workflow') :
      $this->t("Editing '@name' workflow", ['@name' => $this->entity->label()]);

    $form['label'] = [
      '#title' => $this->t('Transition name'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#size' => 40,
      '#maxlength' => 80,
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#maxlength' => 12,
      '#machine_name' => [
        'exists' => '\Drupal\mcapi\Entity\Workflow::load',
      ],
      '#disabled' => !$this->entity->isNew(),
    ];
    $form['summary'] = [
      '#title' => $this->t('Summary'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->summary,
      '#size' => 40,
      '#maxlength' => 80,
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $creation_states = $this->removeState($this->allStates, 'erased');

    $form['creation'] = [
      '#title' => $this->t('Creating the transaction'),
      '#type' => 'fieldset',
      '#tree' => TRUE,
      '#weight' => $w++,
      'state' => [
        '#title' => t('Initial state'),
        '#type' => 'select',
        '#empty_option' => '-'.$this->t('Select').'-',
        '#required' => TRUE,
        '#options' =>  array_combine($creation_states, $creation_states),
        '#default_value' => $this->entity->creation ? $this->entity->creation['state'] : '',
        '#ajax' => [
          'callback' => [$this, 'showStatesFieldsOnly'],
          'wrapper' => 'edit-states',
        ]
      ],
      'confirm' => [
        '#title' => t('Confirm transaction on a separate page'),
        '#type' => 'checkbox',
        '#default_value' => $this->entity->creation ? $this->entity->creation['confirm'] : FALSE
      ]
    ];

    $form['direction'] = [
      '#title' => $this->t('Direction'),
      '#type' => 'radios',
      '#options' => [
        '3rdparty' => $this->t('3rdparty (usually admin only)'),
        'bill' => $this->t('Bill - the user is asking for a payment.'),
        'credit' => $this->t('Credit - the user is sending a payment.'),
      ],
      '#default_value' => $this->entity->direction,
      '#required' => TRUE,
      '#weight' => $w++
    ];
    $form['states'] = [
      '#markup' => '<div id="edit-states"></div>',
      '#weight' => $w++
    ];
    if (isset($this->entity->creation['state'])) {
      $form['states'] = $this->showStatesFields($form, $form_state);
      $form['states']['#weight'] = $w++;
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  function validateForm(array &$form, FormStateInterface $form_state) {
    // No validation required.
  }

  /**
   * {@inheritDoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    parent::save($form, $form_state);
    $form_state->setRedirect('mcapi.admin');
  }

  /**
   * {@inheritdoc}
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $entity = parent::buildEntity($form, $form_state);
    if ($form_state->getValue(['states', 'new_transition'])) {
      [$from_state, $dest_state] = explode('-', $form_state->getValue(['states', 'new_transition']));
      $entity->states[$from_state][$dest_state] = ['actors' => ['payer' => FALSE, 'payee' => FALSE], 'label' => ''];
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    if ($this->entity->local and !$this->entity->isNew()) {
      $actions['delete']['#url']->setOption('query', ['destination' => '/admin/accounting']);
    }
    else {
      unset($actions['delete']);
    }
    return $actions;
  }


  /**
   * Generate the nested states part of the form based on the entity.
   */
  function showStatesFields($form, $form_state) {
    $additional_fields = [
      '#title' => $this->t('Transitions'),
      '#type' => 'details',
      '#open' => TRUE,
      '#tree' => TRUE,
      '#description' => $this->t('Admin can do all defined transitions.'),
    ];
    foreach ($this->entity->states as $from_state => $to_states) {
      $additional_fields[$from_state] = [
        '#title' => $from_state,
        '#type' => 'details',
        '#collapsed' => FALSE
      ];
      foreach ($to_states as $to_state => $info) {
        $additional_fields[$from_state][$to_state] = [
          '#title' => $to_state,
          '#type' => 'details',
          '#open' => TRUE,
          'label' => [
            '#title' => $this->t('Name'),
            '#type' => 'textfield',
            '#required' => TRUE,
            '#default_value' => $info['label'],
            '#weight' => 1
          ],
          'actors' => [
            '#title' => $this->t('Actor'),
            '#type' => 'checkboxes',
            '#required' => TRUE,
            '#options' => [
              'payee' => $this->t('Payee'),
              'payer' => $this->t('Payer')
            ],
            '#default_value' => $info['actors'],
            '#weight' => 2
          ],
        ];
      }
    }
    if ($options = $this->unusedTransitions()) {
      // @todo. Bad selection when making new type.
      $additional_fields['new_transition'] = [
        '#title' => $this->t('Add transition'),
        '#type' => 'select',
        '#empty_option' => ' - '.$this->t('Select').' - ',
        '#default_value' => '',
        '#options' => $options,
        '#ajax' => [
          'callback' => [$this, 'showStatesFieldsOnly'],
          'wrapper' => 'edit-states',
        ]
      ];
    }
    return $additional_fields;
  }

  /**
   * Ajax callback
   * @param type $form
   * @param type $form_state
   * @return array
   */
  function showStatesFieldsOnly($form, $form_state) {
    return $form['states'];
  }

  /**
   * Get a list of possible, but unused transitions.
   * @return array
   */
  function unusedTransitions() : array {
    $transitions = [];
    if (empty($this->entity->creation['state'])) {

    }
    if (empty($this->entity->states)) {
      $transitions[$this->entity->creation['state']] = $this->removeState($this->allStates, $this->entity->creation['state']);
    }
    else {
      $possible_transitions = array_map(function ($a){return $this->removeState($this->allStates, $a) ;}, $this->allStates);
      foreach ($this->entity->states as $from_state => $dest_states) {
        foreach ($dest_states as $dest_state => $info) {
          if (isset($this->entity->states[$from_state][$dest_state])) {
            unset($possible_transitions[$from_state][$dest_state]);
          }
        }
      }
      $transitions = $possible_transitions;
    }
    $options = [];
    foreach ($transitions as $start_state => $end_states) {
      foreach ($end_states as $end_state) {
        $options["$start_state-$end_state"] =  "$start_state -> $end_state";
      }
    }
    return $options;
  }

  /**
   * Generate the part of the form which chooses a new dest state given a new from state
   */
  function addTransition($form, $form_state) {
    [$from_state, $dest_state] = explode('-', $form_state->getValue('states')['new_transition']);
    $form['states'][$from_state][$dest_state] = [
      '#title' => $dest_state,
      '#type' => 'details',
      '#weight' => 7,
      'to_state' => [
        '#title' => $this->t("Add 'dest' state"),
        '#type' => 'select',
        '#required' => TRUE,
        '#empty_option' => '-'.$this->t('Select').'-',
        '#options' => $choices,
        '#ajax' => [
          'callback' => [$this, 'addDestState'],
          'wrapper' => "edit-states",
        ]
      ]
    ];
    return $form['states'];
  }

  /**
   * Remove a state froim a given list of states.
   * @param array $states
   * @param string $state
   * @return array
   */
  protected function removeState(array $states, $state) : array {
    return array_diff($states, [$state]);
  }

}

