<?php

namespace Drupal\Mcapi;

use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Component\Utility\Html;

/**
 * Utility class for community accounting.
 */
class Functions {

  /**
   * Get a list of entity labels keyed by entity ID suitable for a select box.
   * @param string $entity_type_id
   * @return string[]
   *
   * @note this is only used for mc_state, mc_workflow ATM
   */
  static function entityOptions(string $entity_type_id) : array {
    $entities = \Drupal::entityTypeManager()->getStorage($entity_type_id)->loadMultiple();
    $contextRepo = \Drupal::service('entity.repository');
    foreach ($entities as $entity) {
      $label = $contextRepo->getTranslationFromContext($entity)->label();
      $list[$entity->id()] = Html::escape($label);
    }
    return $list;
  }

  /**
   * Get a list of all the tokens on the transaction entity.
   *
   * @note This really needs replacing with a core function.
   */
  public static function tokenHelp() : string {
    foreach (array_keys(\Drupal::Token()->getInfo()['tokens']['xaction']) as $token) {
      $tokens[] = "[xaction:$token]";
    }
    return implode(', ', $tokens);
  }
  /**
   * Convert the raw value into numbers for a main widget and subdivision widget.
   */
  public static function formatCurrencyParts(int $raw_num) : array {
    [,$main_format,,$subdivision_format] = array_values(\Drupal::Config('mcapi.settings')->get('currency.format'));
    $formatted = ['main' => $raw_num, 'subdivision' => NULL];
    if ($subdivision_format === NULL) {
      if ($pos = strrpos($main_format, '.')) {
        $formatted['main'] *= pow(10, -$pos);
      }
    }
    else {
      list($subs) = explode('/', $subdivision_format);
      $divisor = (int)$subs + 1; // number of subs in a single unit e.g. 100
      $formatted['main'] = intval(floor($raw_num / $divisor));
      $formatted['subdivision'] = $raw_num % $divisor;
    }
    return $formatted;
  }

  /**
   * @param int $raw_num
   * @return Markup|float
   */
  public static function formatCurrency($raw_num, $linked = TRUE) : Markup {
    // Replace the parts back into the number format.
    $output = implode(static::formatCurrencyHelper($raw_num));
    // Link it to the currency stats page.
    if ($linked) {
      $output = (string)Link::createFromRoute(['#markup' => $output], 'mcapi.metrics')->toString();
    }
    return Markup::create($output);
  }

  /**
   * Format the value as a machine-readable number
   *
   * @param type $raw_num
   * @return float
   */
  public static function formatCurrencyAsFloat($raw_num) : float {
    $output = static::formatCurrencyHelper($raw_num);
    // replace fractions with decimals
    if (strpos($output['subdivision'],';')) {
      $output = str_replace(
        ['&#189;', '&#8531', '&#8532', '&#188;', '&#8533;', '&#8534;', '&#8535;', '&#8536;'],
        ['.5', '.33', '.66', '.25','.2', '.4', '.6', '.8'],
        $output
      );
    }
    elseif ($output['subdivision']) { // convert to decimal
      $format = \Drupal::config('mcapi.settings')->get('currency.format');
      $divisions = substr($format['subdivision'], 0, strpos($format['subdivision'], '/')); //e.g. 59 mins
      $powerOfTen = strlen($divisions);
      $denominator = pow(10, $powerOfTen);
      $output['subdivision'] = round($denominator / ($divisions+1), $powerOfTen);
    }
    $output = implode($output);
    // Convert 1hr 23mins to 1.23 using regex
    // Replace likely separators with decimal point dots
    $output = preg_replace('/([.,:]+)/', '.', strip_tags($output));
    // Remove everything except numbers and dots and the minus sign.
    $output = preg_replace('/[^0-9.-]/', '', $output);
    return (float)$output;
  }

  /**
   * Prepare to format a currency value.
   * @param int $raw_num
   * @return array
   */
  private static function formatCurrencyHelper($raw_num) : array {
    $format = $output = \Drupal::config('mcapi.settings')->get('currency.format');
    foreach (self::formatCurrencyParts(abs($raw_num)) as $key => $num) {
      $output[$key] = $num;
    }
    if (!empty($format['subdivision'])) {
      $pos = strpos($format['subdivision'], '/');
      if ($output['subdivision']) {
        if ($pos) { // specific subdivisions allowed
          // Don't use / in the pad length. ⅜
          $decimal = $output['subdivision'] / intval(substr($format['subdivision'], 0, $pos)+1);
          if ($frac = static::decToFraction($decimal + 1)) {
            // Frac is a unicode string denoting a fraction
            $output['f2'] = '';
            $output['subdivision'] = $frac;
          }
        }
      }
      else {
        $output['subdivision'] = str_pad($output['subdivision'], $pos, '0', STR_PAD_LEFT);
      }
    }
    // Subdivision is a decimal in the main field
    elseif ($pos = strpos($format['main'], '.')) {
      $dps = strlen($format['main']) - ($pos+1);
      $output['main'] = number_format((float)$output['main'], $dps, '.', '');
    }
    if ($raw_num < 0) {
      // always render it as a positive number.
      $output['f1'] = '-'.$output['f1'];
    }
    return $output;
  }

  /**
   * @param type $float
   * @return string
   * @kudos https://stackoverflow.com/questions/1954018/php-convert-decimal-into-fraction-and-back#9143510
   * @kudos https://www.alt-codes.net/fraction-symbols
   */
  private static function decToFraction($float) : string | null {
    // 1/2, 1/4, 1/8, 1/16, 1/3 ,2/3, 3/4, 3/8, 5/8, 7/8, 3/16, 5/16, 7/16,
    // 9/16, 11/16, 13/16, 15/16
    $whole = floor ( $float );
    $decimal = $float - $whole;
    $leastCommonDenom = 48; // 16 * 3;
    $denominators = array (2, 3, 4, 5);
    $roundedDecimal = round ( $decimal * $leastCommonDenom ) / $leastCommonDenom;
    if ($roundedDecimal == 0)
      return $whole;
    if ($roundedDecimal == 1)
      return $whole + 1;
    foreach ($denominators as $d ) {
      if ($roundedDecimal * $d == floor($roundedDecimal * $d )) {
        $denom = $d;
        break;
      }
    }
    if (!isset($denom)) {
      return NULL;
    }

    $numerator = intval($denom * $roundedDecimal);
    if ($denom) {
      switch ($denom) {
        case 2:
          $frac = '&#189;';
          break;
        case 3:
          if ($numerator == 1)$frac = '&#8531';
          else $frac = '&#8532';
          break;
        case 4:
          if ($numerator == 1)$frac = '&#188;';
          else $frac = '&#190;';
          break;
        case 5:
          if ($numerator == 1)$frac = '&#8533;';
          elseif ($numerator == 2)$frac = '&#8534;';
          elseif ($numerator == 3)$frac = '&#8535;';
          elseif ($numerator == 4)$frac = '&#8536;';
          break;
        default:
          //$frac = ($whole == 0 ? '' : $whole) . " " . ($roundedDecimal * $denom) . "/" . $denom;
      }
      return $frac;
    }

  }


  /**
   * Get the transactions needing signature of the given user.
   *
   * @param int $uid
   *   The user ID.
   *
   * @return int[]
   *   Transaction IDs
   */
  static function transactionsNeedingSigOfUser($uid) : array {
    return \Drupal::entityQuery('mc_transaction')
      ->condition('signatures.uid', $uid)
      ->condition('signatures.signed', 0)
      ->accessCheck(FALSE)
      ->execute();
  }

  /**
   * Transactions which the user has signed but are still pending.
   * @param int $uid
   * @return int[]
   *   Transaction IDs
   */
  static function transactionsNeedingSig(int $uid) : array {
    return \Drupal::entityQuery('mc_transaction')
      ->condition('signatures.uid', $uid)
      ->condition('state', 'pending')
      ->accessCheck(FALSE)
      ->execute();
  }

  /**
   * Assembles transaction and returns the transaction with added entries in its
   * creation->state. To ensure the transaction is built only once, this stores
   * it as a static. This means it can be used to alter a passed transaction or
   * returns a previously passed transaction.
   *
   * @staticvar type $assembled
   * @param TransactionInterface $transaction
   * @return TransactionInterface
   *
   * @note this method cannot belong in the transaction object.
   */
  static function getAssembledTransaction(TransactionInterface $transaction = NULL) : TransactionInterface {
    static $assembled = [];
    $uuid = $transaction->uuid();
    if (!$transaction->isNew()) {
       throw new \Exception ('Called Functions::getAssembledTransaction() with a non-new transaction.');
    }
    if (isset($assembled[$uuid])) {
      $transaction = $assembled[$uuid];
    }
    else {
      // Set the transaction to its initial state and add signatories.
      // Allow other modules to add entries, but not to modify the transaction.
      $entries = \Drupal::moduleHandler()->invokeAll('transaction_assemble', [clone($transaction)]);
      foreach ($entries as $params) {
        $params['primary'] = FALSE;
        $transaction->addEntry($params);
      }
      $assembled[$uuid] = $transaction;
    }
    return $transaction;
  }

  /**
   * Utility
   * @param Transaction $transaction
   * @return []
   *   each value is an array with title and Url
   */
  static function getTransactionOperations(TransactionInterface $transaction) {
    $links = [];
    if (!$transaction->isNew()) {
      $role = '';
      // derive the other links from the workflow.
      $current_user = \Drupal::currentUser();
      if ($transaction->payee->getOwnerId() == $current_user->id()) {
        $role = 'payee';
      }
      elseif ($transaction->payer->getOwnerId() == $current_user->id()) {
        $role = 'payer';
      }
      // Drupal cannot be admin for a remote transaction.
      foreach ($transaction->workflow->entity->states[$transaction->state->value] as $state_id => $info) {
        if (in_array($role, $info['actors']) or $current_user->hasPermission('manage mcapi')) {
          $redirect_route = ($state_id == 'delete') ? 'entity.mc_transaction.collection' : '<current>';
          $links[$state_id] = [
            'title' => $info['label'],
            'url' => \Drupal\Core\Url::fromRoute(
              'entity.mc_transaction.transition',
              ['mc_transaction' => $transaction->id(), 'dest_state' => $state_id],
              ['query' => ['destination' => \Drupal\Core\Url::fromRoute($redirect_route)->toString()]]
            )
          ];
        }
      }
    }
    return $links;
  }
}