<?php

namespace Drupal\mcapi\Plugin\DevelGenerate;

use Drupal\mcapi\Entity\Workflow;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Wallet;
use Drupal\devel_generate\DevelGenerateBase;
use Drupal\user\Entity\User;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Extension\ExtensionPathResolver;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Component\Datetime\Time;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a DevelGenerate plugin.
 *
 * @DevelGenerate(
 *   id = "mc_transaction",
 *   label = @Translation("transactions"),
 *   description = @Translation("Generate a given number of transactions..."),
 *   url = "transaction",
 *   permission = "administer devel_generate",
 *   settings = {
 *     "num" = 100,
 *     "kill" = TRUE,
 *     "type" = "default"
 *   }
 * )
 * account-switcher, logger, currentuser, time, moduleHandler, extension.path.resolver
 */
class TransactionDevelGenerate extends DevelGenerateBase implements ContainerFactoryPluginInterface {

  const MAX = 100;

  /**
   * The transaction storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $transactionStorage;

  /**
   * @var ExtensionPathResolver
   */
  protected $extensionPathResolver;

  /**
   * @var AccountSwitcherInterface
   */
  protected $accountSwitcher;

  /**
   * @var LoggerChannel
   */
  protected $logger;

  /**
   * @var AccountProxyInterface
   */
  protected $currentUser;

  /**
   * @var ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The unixtime of the request
   * @var Int
   */
  protected $requestTime;

  /**
   * All the wallet ids
   *
   * @var array
   */
  private array $wids;

  /**
   * @param array $configuration
   * @param type $plugin_id
   * @param array $plugin_definition
   * @param EntityTypeManagerInterface $entity_type_manager
   * @param ExtensionPathResolver $extension_path_resolver
   * @param AccountSwitcherInterface $account_switcher
   * @param LoggerChannelFactoryInterface $logger_factory
   * @param AccountProxyInterface $current_user
   * @param ModuleHandlerInterface $module_handler
   * @param Time $time
   * @throws \Exception
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    array $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    ExtensionPathResolver $extension_path_resolver,
    AccountSwitcherInterface $account_switcher,
    LoggerChannelFactoryInterface $logger_factory,
    AccountProxyInterface $current_user,
    ModuleHandlerInterface $module_handler,
    Time $time
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->transactionStorage = $entity_type_manager->getStorage('mc_transaction');
    $this->extensionPathResolver = $extension_path_resolver;
    $this->accountSwitcher = $account_switcher;
    $this->logger = $logger_factory->get('devel_generate');
    $this->currentUser = $current_user;
    $this->moduleHandler = $module_handler;
    $this->wids = $entity_type_manager->getstorage('mc_wallet')->getQuery()->accessCheck(TRUE)->execute();
    if (count($this->wids) < 2) {
      throw new \Exception('Not enough wallets to make a transaction.');
    }
    $this->requestTime = $time->getRequestTime();
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration, $plugin_id, $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('extension.path.resolver'),
      $container->get('account_switcher'),
      $container->get('logger.factory'),
      $container->get('current_user'),
      $container->get('module_handler'),
      $container->get('datetime.time'),
    );
  }

  /**
   * {@inheritDoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form['kill'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Delete existing transactions.'),
      '#default_value' => $this->getSetting('kill'),
    ];
    $form['num'] = [
      '#type' => 'number',
      '#title' => $this->t('How many transactions would you like to generate?'),
      '#default_value' => $this->getSetting('num'),
      '#required' => TRUE,
      '#min' => 0,
    ];
    $form['#redirect'] = FALSE;
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function generateElements(array $values) {
    $this->settings = $values + $this->settings;
    if ($this->getSetting('num') <= static::MAX) {
      if ($this->getSetting('kill')) {
        $this->contentKill();
      }
      $workflows = Workflow::loadMultiple();
      for ($i = 0; $i < $this->getSetting('num'); $i++) {
        $this->develGenerateTransactionAdd(array_rand($workflows));
      }
      if (function_exists('drush_log') && $i % drush_get_option('feedback', 1000) == 0) {
        drush_log(dt('Completed @feedback transactions ', ['@feedback' => drush_get_option('feedback', 1000)], 'ok'));
      }
      $this->setMessage($this->t('Created @num transactions.', ['@num' => $this->getSetting('num')]));
    }
    else {
      // setFile doesn't include the file yet and devel_generate_batch_finished must be callable
      $this->moduleHandler->loadInclude('devel_generate', 'batch.inc');
      $batch_builder = (new BatchBuilder())
       ->setTitle($this->t('Generating transactions'))
       ->setFile($this->extensionPathResolver->getPath('module', 'devel_generate') . '/devel_generate.batch.inc')
       ->setFinishCallback('devel_generate_batch_finished');
      // Add the kill operation.
      if ($this->getSetting('kill')) {
        $batch_builder->addOperation('devel_generate_operation', [$this, 'batchContentKill', []]);
      }

      $batches = ceil($this->getSetting('num')/static::MAX);
      for ($num = 0; $num < $batches; $num++) {
        $batch_builder->addOperation('devel_generate_operation', [$this, 'batchContentAddTransactions', []]);
      }
      batch_set($batch_builder->toArray());
    }
  }

  /**
   * batch callback to create one transaction.
   */
  public function batchContentAddTransactions($values, &$context) {
    $i = 0;
    if (!isset($context['results']['num'])) {
      $context['results']['num'] = 0;
    }

    $this->wids = $this->getEntityTypeManager()
      ->getstorage('mc_wallet')
      ->getQuery()->accessCheck(TRUE)
      ->execute();
    $workflows = Workflow::loadMultiple();
    while ($context['results']['num'] < $this->getSetting('num') and $i < Self::MAX) {
      $this->develGenerateTransactionAdd(array_rand($workflows));
      $context['results']['num']++;
      $i++;
    }
  }

  /**
   * batch callback to create one transaction.
   */
  public function batchContentKill($values, &$context) {
    $this->contentKill();
  }

  /**
   * Deletes all transactions .
   *
   * @param string $workflow
   *   The type of transactions to delete
   *
   * @note Loads all transactions into memory at the same time.
   */
  public function contentKill() {
    $xids = $this->transactionStorage->getQuery()->accessCheck(FALSE)->execute();
    if (!empty($xids)) {
      $transactions = Transaction::loadMultiple($xids);
      $this->transactionStorage->delete($transactions);
      $this->setMessage($this->t('Deleted %count transactions.', array('%count' => count($xids))));
    }
  }

  /**
   * Create one transaction. Used by both batch and non-batch code branches.
   *
   * Ideally this would save multiple versions of transactions, progressing through a real workflow
   */
  public function develGenerateTransactionAdd($workflow_id) {
    list($w1, $w2) = $this->get2RandWalletIds();
    if (!$w2) {
      return;
    }
    $creator = Wallet::load($w1)->getOwner();
    $props = [
      'payer' => $w1,
      'payee' => $w2,
      // Transactions of type 'inherit' don't show in the default view.
      'workflow' => $workflow_id,
      'creator' => $creator->id(),
      'description' => $this->getRandom()->sentences(1),
      'quant' => rand(10, 250)
    ];
    $transaction = Transaction::create($props);
    // Populate additional fields
    $this->populateFields($transaction);
    // We don't want all transactions created at the same time!
    // Make created time between the newest involved wallet creation and now.
    $latest = 0;
    foreach (Wallet::loadMultiple($transaction->allWalletIds()) as $wallet) {
      $latest = max($latest, $wallet->created->value);
    }
    $transaction->created->value = rand($latest, $this->requestTime);
    $workflow = $transaction->workflow->entity;
    // Pretend to be the payer or payee
    $use_as = NULL;
    if ($workflow->direction == 'credit') {
      $use_as = $transaction->entries->payer->getOwner();
    }
    elseif ($workflow->direction == 'bill') {
      $use_as = $transaction->entries->payee->getOwner();
    }
    try {
      if (isset($use_as)) {
        $this->accountSwitcher->switchTo($use_as);
      }
      // Assemble the transaction.
      $violations = $transaction->validate();
      if ($violations->count()) return;// probably balance limits
    }
    catch (\Exception $e) {
      $this->logger->error(
        'Transaction failed validation: @message\n@transaction',
        ['@message' => $e->getMessage(), '@transaction' => print_r($transaction, 1)]
      );
      return;
    }
    finally {
      if (isset($use_as)) {
        $this->accountSwitcher->switchback();
      }
    }
    // Saving the first version
    $this->saveTransactionAsUid($transaction, $creator->id());
    if (empty($workflow->states)) return;
    while (TRUE) {
      $rand = rand(0, 99);
      switch ($transaction->state->value) {
        case 'pending':
          if ($rand > 95) {
            return;
          }
          if ($uids = $transaction->signatures->getUnsigned()) {
            $uid = reset($uids);
            $transaction->signatures->sign($uid);
            $next_state = $transaction->signatures->getUnsigned() ? 'pending' : 'completed';
            $transaction->newVersion($next_state);
            $this->saveTransactionAsUid($transaction, $uid);
          }
          break;

        case 'completed':
          if ($rand < 97) {
            return;
          }
          $next_states = $workflow->states[$transaction->state->value];
          $next_state = array_rand($next_states);
          $transaction->newVersion($next_state);
          $info = $next_states[$next_state];
          $actors = array_filter($info['actors'] + ['admin']);
          shuffle($actors);
          $actor = array_pop($actors);
          if ($actor == 'admin')$uid = $this->currentUser->id();
          elseif ($actor == 'payee')$uid = $transaction->payee->getOwnerId();
          elseif ($actor == 'payer')$uid = $transaction->payer->getOwnerId();
          $this->saveTransactionAsUid($transaction, $uid);
          break;

        case 'erased':
          return;

        default:
          die('Illegal state in TransactionDevelGenerate: '.$transaction->state->value);
      }
    }
  }

  private function saveTransactionAsUid($transaction, $uid) {
    $user = User::load($uid);
    $this->accountSwitcher->switchTo($user);
    $transaction->save();
    $this->accountSwitcher->switchBack();
  }

  /**
   * Get two random wallets
   *
   * @return int[]
   *   2 wallet ids
   */
  public function get2RandWalletIds() {
    $wids = $this->wids;
    shuffle($wids);
    return array_slice($wids, -2);
  }

  /**
   * {@inheritDoc}
   */
  public function validateDrushParams($args, $options = []) {
    $values['kill'] = drush_get_option('kill');
    $values['num'] = array_shift($args);
    return $values;
  }

}
